package agents;

import java.util.Calendar;
import java.util.ArrayList;
import java.util.Random;
import java.util.UUID;

import ontology.Driver;
import ontology.LogJourneyData;
import ontology.OccupationRequest;
import ontology.ParkingManagementOntology;
import ontology.Reservation;
import ontology.ReservationRequest;
import ontology.Road;
import ontology.UpdateTime;
import ontology.Vehicle;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;

/**
 * @author Francisco Palau Romero
 * @code: Journey Manager agent code 
 *
 */

public class AgJourneyManager extends Agent {
	//Vehicle parameters
	private Road road;
	private int km, km_in, km_out; //Kms
	private int drivingSpeed = 90; //Km/h
	private String journeyManagerId;
	private int vehicleId;
	private int restTime = 60; //Minutes
	private int maximunDriveTime = 270; //Minutes
	private int drivedTime = 0; //Minutes
	private int drivedTimeBeforeRest = 0; //Minutes
	private int lostTime = 0; //Minutes
	private int restedTime = 0; //Minutes
	//Driver parameters
	private Driver driver;
	//General parameters
	private AID ID;
	private Calendar actualTime = Calendar.getInstance();
	private Calendar estimatedArrivalTime = Calendar.getInstance();
	private Calendar estimatedEndingTime = Calendar.getInstance();
	private int infractionCommited = 0;
	private Calendar timeIn = Calendar.getInstance();
	private Calendar timeOut = Calendar.getInstance();
	//Simulation type
	private int simulationType;
	private static final int BORDA_NEGOTIATION_SIMULATION_TYPE = 0;
	private static final int FREE_SIMULATION_TYPE = 1;
	//Desired rest areas list
	private ArrayList<AID> RestAreaList = new ArrayList<AID>();
	private ArrayList<Float> RestAreaDistanceList = new ArrayList<Float>();
	private ArrayList<Float> RestAreaKMList = new ArrayList<Float>();
	boolean listCompleted = false;
	boolean creatingList = false;
	private AID nextRestArea;
	private boolean reachRestArea = false; //For free reservation mode only
	private boolean resting = false;
	//Ontology
	private Codec codec = new SLCodec();
	private Ontology ontoloy = ParkingManagementOntology.getInstance();
	private Action action;
	//Log AID
	private AID logManagerAID;

	//******************************************************************************************//
	//******************************************************************************************//
	// Setup																					//
	//******************************************************************************************//
	//******************************************************************************************//
	protected void setup() {
		//Register codec and ontology
	    getContentManager().registerLanguage(codec);
	    getContentManager().registerOntology(ontoloy);
	    //

	    //Extract parameters from Loader
		Object[] args = getArguments();
		Vehicle vehicleInfo = (Vehicle) args[0];
		
		road = vehicleInfo.getRoad();
		km = vehicleInfo.getKM();
		km_in = vehicleInfo.getKM();
		//Calculating the km out (value between the km in which the truck is and the maximum length of the road)
		int km_max = road.getLenght();
		Random myRandomizer = new Random();
		km_out = myRandomizer.nextInt((km_max - km) + 1) + km;
		//
		drivingSpeed = vehicleInfo.getMaxSpeed();
		vehicleId = vehicleInfo.getVehicleId();
		journeyManagerId = getLocalName();
		driver = vehicleInfo.getDriver();
		drivedTime = vehicleInfo.getDriver().getDrivingTime();
		drivedTimeBeforeRest = drivedTime;

		simulationType = (Integer) args[1];
		actualTime = (Calendar) args[2];
		timeIn.setTime(actualTime.getTime());
		//
		

		//Register the offered service
		DFAgentDescription dfd = new DFAgentDescription();
		ID = getAID();
		dfd.setName(ID);
		ServiceDescription sd = new ServiceDescription();
		sd.setType("JourneyManager");
		sd.setName("truck"+vehicleId);
		dfd.addServices(sd);
		try{
			DFService.register(this, dfd);
		} catch (Exception e) {e.printStackTrace();}
		//

		//Get the AIDs of the log
		DFAgentDescription[] result = null;
		ServiceDescription sdLog = new ServiceDescription(); 
		DFAgentDescription dfdLog = new DFAgentDescription();

		sdLog.setType("LogManager");
		sdLog.setName("logManager");
		dfdLog.addServices(sdLog);

		while (result == null) {
		    try {
		    	//Search for 100 ms
		    	result = DFService.searchUntilFound(this, getDefaultDF(), dfdLog, null, 100);
			} catch (FIPAException e) { 
				e.printStackTrace(); 
				System.out.println("FIPA Error while searching for agents");
			}
		}
	    if (result.length >= 1) {
	    	for (DFAgentDescription logManagerResult : result){
	    		logManagerAID = logManagerResult.getName();
	    	}
	    }
	    //
	    
	    //The type of simulation it will depend of simulationType
	    switch (simulationType) {
		    case FREE_SIMULATION_TYPE:
				addBehaviour(new BJourneyManagerReservationFree());
				addBehaviour(new BJourneyManagerUpdateFree());
		    	break;
		    case BORDA_NEGOTIATION_SIMULATION_TYPE:
				addBehaviour(new BJourneyManagerReservation());
				addBehaviour(new BJourneyManagerNegotiation());
				addBehaviour(new BJourneyManagerUpdate());
		    	break;
	    }
	    //

		//Shut down behaviour
		addBehaviour(new BJourneyManagerShutDown());
		//
	}
	//******************************************************************************************//
	//******************************************************************************************//

	
	//******************************************************************************************//
	//******************************************************************************************//
	// takeDown: Sending logs															//
	//******************************************************************************************//
	//******************************************************************************************//
	@Override
	protected void takeDown(){
		//TimeOut
		timeOut.setTime(actualTime.getTime());
		//Send the log with the journey data to logManager
		ACLMessage msgLogJourneyData = new ACLMessage(ACLMessage.INFORM);
		msgLogJourneyData.setLanguage(codec.getName());
		msgLogJourneyData.setOntology(ontoloy.getName());
		LogJourneyData logJourneyData = new LogJourneyData();
		logJourneyData.setIdVehicle(vehicleId);
		logJourneyData.setdrivedTime(drivedTime);
		logJourneyData.setKm(km_out - km_in);
		logJourneyData.setKmIn(km_in);
		logJourneyData.setKmOut(km_out);
		logJourneyData.setlostTime(lostTime);
		logJourneyData.setrestTime(restedTime);
		logJourneyData.setTimeIn(timeIn.getTime());
		logJourneyData.setTimeOut(timeOut.getTime());
		logJourneyData.setInfractionType(infractionCommited);
		//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
		action = new Action(getAID(),logJourneyData);
		try {
			getContentManager().fillContent(msgLogJourneyData, action);
		} catch (CodecException | OntologyException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		msgLogJourneyData.setConversationId("log");
		msgLogJourneyData.addReceiver(logManagerAID);
		send(msgLogJourneyData);
		//

		while (receive() != null) {};
	}
	//******************************************************************************************//
	//******************************************************************************************//

	
	//******************************************************************************************//
	//******************************************************************************************//
	// Borda Negotiation Behavious: BJourneyManagerReservation, BJourneyManagerNegotiation,		//
	//								BJourneyUpdate												//
	//******************************************************************************************//
	//******************************************************************************************//
	private class BJourneyManagerReservation extends Behaviour {

		int step = 0;
		MessageTemplate mtRequestReply;
		AID bestRestArea;
		String reservationCode;
		int bestIndex = 0;
		int index = 0;
		float bestDistance;
		ArrayList<AID> newRestAreaList = new ArrayList<AID>();
		ArrayList<Float> newRestAreaDistanceList = new ArrayList<Float>();

		@Override
		public void action() {
			switch (step){
			//Creating the Preference List
			case 0:
				if (!creatingList)	{
					addBehaviour(new BJourneyManagerCreatingPreferencesTable());
					creatingList = true;
				}
				if (listCompleted) {
					step++;
					creatingList = false;
					newRestAreaList.addAll(RestAreaList);
					newRestAreaDistanceList.addAll(RestAreaDistanceList);
				}
				break;
			//Choosing the best rest area to park (it is the rest area that allows the truck drive the maximun distance
			case 1:
				//If the journey manager runs out of options, it must exit of the system
				if (newRestAreaList.isEmpty()){
					infractionCommited = 1;
					myAgent.doDelete();
					step = 3;
					break;
				}
				bestIndex = 0;
				index = 0;
				bestDistance = 0;
				//The best will be the one most far away from my position (the one which is value is greater)
				for (float distance : newRestAreaDistanceList){
					if (distance > bestDistance) {
						bestIndex = index;
						bestDistance = distance;
					}
					index++;
				}
				bestRestArea = newRestAreaList.get(bestIndex);
				//For the moment, we will assume that we got a reservation there, just in case there is a negotiation
				nextRestArea = bestRestArea;
				//

				ACLMessage msgRequestParkingSpot = new ACLMessage(ACLMessage.REQUEST);
				reservationCode = UUID.randomUUID().toString();

				msgRequestParkingSpot.addReceiver(bestRestArea);
				msgRequestParkingSpot.setConversationId("request");
				
				//Creating the reservation request message using ontology
				msgRequestParkingSpot.setLanguage(codec.getName());
				msgRequestParkingSpot.setOntology(ontoloy.getName());
				Reservation reservation = new Reservation();
				Calendar timeStart, timeEnd;
				timeStart = Calendar.getInstance();
				timeEnd = Calendar.getInstance();
				timeStart.add(Calendar.MINUTE,(int) (bestDistance*60/drivingSpeed));
				estimatedArrivalTime.setTime(timeStart.getTime());
				timeEnd.add(Calendar.MINUTE, (int) (bestDistance*60/drivingSpeed)+restTime);
				estimatedEndingTime.setTime(timeEnd.getTime());
				reservation.setTimeStart(timeStart.getTime());
				reservation.setTimeEnd(timeEnd.getTime());
				reservation.setjourneyManagerId(journeyManagerId);
				ReservationRequest reservationRequest = new ReservationRequest();
				reservationRequest.setReservation(reservation);
				//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
				action = new Action(getAID(),reservationRequest);
				try {
					getContentManager().fillContent(msgRequestParkingSpot, action);
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				msgRequestParkingSpot.setInReplyTo(reservationCode);
				
				//Send the reservation
				myAgent.send(msgRequestParkingSpot);
				step++;
				break;
			case 2:

				//Waiting for response
				mtRequestReply = MessageTemplate.MatchInReplyTo(reservationCode);

				ACLMessage msgCFPReply = myAgent.receive(mtRequestReply);
				
				if (msgCFPReply != null) {

					//If everything is OK, the behavior ends
					if (msgCFPReply.getPerformative()==ACLMessage.AGREE) {
						//The Journey Manager stored the reservation start/end and target rest area, so this behaviour ends
						step++;
						break;
					}
					//If there was a problem, the next best rest area will be selected
					if (msgCFPReply.getPerformative()==ACLMessage.REFUSE) {
						nextRestArea = null;
						newRestAreaList.remove(bestIndex);
						newRestAreaDistanceList.remove(bestIndex);
						step--;
						break;
					}
				} else block();
				break;
			}
		}

		@Override
		public boolean done() {
			if (step == 3) return true;
			return false;
		}
	}



	//This behavior begins with an CFP message, and the journey manager must offer another option (rest area)
	private class BJourneyManagerNegotiation extends Behaviour {
		int step = 0;
		MessageTemplate mtCFP = MessageTemplate.MatchPerformative(ACLMessage.CFP);
		MessageTemplate mtCFPAccept;
		String replyCode;
		AID restAreaManager;
		int index, bestIndex;
		String reservationCode;
		ArrayList<AID> newRestAreaList = new ArrayList<AID>();
		ArrayList<Float> newRestAreaDistanceList = new ArrayList<Float>();
		ArrayList<Float> newRestAreaKMList = new ArrayList<Float>();
		String ocupationCode;
		Reservation reservation;
		OccupationRequest occupationRequest;
		ReservationRequest reservationRequest;

		@Override
		public void action() {
			switch (step){
			case 0:
				//We recieve a calling for proposal
				ACLMessage msg = myAgent.receive(mtCFP);
				if (msg!=null){
					replyCode = msg.getInReplyTo();
					restAreaManager = msg.getSender();
					newRestAreaList.addAll(RestAreaList);
					newRestAreaDistanceList.addAll(RestAreaDistanceList);
					newRestAreaKMList.addAll(RestAreaKMList);

					step++;
				} else block();
				break;
			case 1:
				//If the vehicle is 15 minutes away from the source of negotiation, the negotiation ends
				float distance = newRestAreaDistanceList.get(newRestAreaList.indexOf(restAreaManager));
				float timeUntilArrival = (distance/drivingSpeed)*60; //Minutes until the vehicle arrives to the rest area
				if (timeUntilArrival <= 15) {
					ACLMessage msgCFPRefuse = new ACLMessage(ACLMessage.REFUSE);
					msgCFPRefuse.addReceiver(restAreaManager);
					msgCFPRefuse.setInReplyTo(replyCode);
					myAgent.send(msgCFPRefuse);
					step = 6;
					break;
				}
				step++;
				break;
			case 2:
				//Is the list is empty, refuse the negotiation
				if (newRestAreaList.isEmpty()) {
					ACLMessage msgCFPRefuse = new ACLMessage(ACLMessage.REFUSE);
					msgCFPRefuse.addReceiver(restAreaManager);
					msgCFPRefuse.setInReplyTo(replyCode);
					myAgent.send(msgCFPRefuse);
					step = 6;
					break;
				}
				//Searching for the best rest area (besides the one which initiate the negotiation)
				distance = 0;
				index = 0;
				bestIndex = -1;
				for (AID restArea : newRestAreaList) {
					if (restArea != restAreaManager && newRestAreaKMList.get(index) >= km && newRestAreaDistanceList.get(index) > distance) {
						distance = newRestAreaDistanceList.get(index);
						bestIndex = index;
					}
					index++;
				}

				//If there is no one to negotiate, refuse the negotiation
				if (bestIndex == -1) {
					ACLMessage msgCFPRefuse = new ACLMessage(ACLMessage.REFUSE);
					msgCFPRefuse.addReceiver(restAreaManager);
					msgCFPRefuse.setInReplyTo(replyCode);
					myAgent.send(msgCFPRefuse);
					step = 6;
					break;
				}
				
				//Ask to the selected rest area if a spot is avaliable
				ocupationCode = UUID.randomUUID().toString();
				
				ACLMessage msgRequestOcupation = new ACLMessage(ACLMessage.REQUEST);
				msgRequestOcupation.addReceiver(newRestAreaList.get(bestIndex));
				msgRequestOcupation.setInReplyTo(ocupationCode);
				msgRequestOcupation.setConversationId("occupation");
				
				//Creating the occupation request message using ontology
				msgRequestOcupation.setLanguage(codec.getName());
				msgRequestOcupation.setOntology(ontoloy.getName());
				reservation = new Reservation();
				Calendar timeStart, timeEnd;
				timeStart = Calendar.getInstance();
				timeEnd = Calendar.getInstance();
				timeStart.add(Calendar.MINUTE,(int) (distance*60/drivingSpeed));
				timeEnd.add(Calendar.MINUTE,(int) (distance*60/drivingSpeed)+restTime);
				reservation.setTimeStart(timeStart.getTime());
				reservation.setTimeEnd(timeEnd.getTime());
				reservation.setjourneyManagerId(journeyManagerId);
				occupationRequest = new OccupationRequest();
				occupationRequest.setReservation(reservation);
				//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
				action = new Action(getAID(),occupationRequest);
				try {
					getContentManager().fillContent(msgRequestOcupation, action);
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

				myAgent.send(msgRequestOcupation);
				step++;
				break;
			case 3:
				//Check if there is free space in the selected rest area
				MessageTemplate mtReplyOccupation = MessageTemplate.MatchInReplyTo(ocupationCode);

				ACLMessage msgReplyOccupation = myAgent.receive(mtReplyOccupation);
				
				if (msgReplyOccupation != null) {
					if (msgReplyOccupation.getPerformative()==ACLMessage.AGREE) {
						ACLMessage msgCFPPropose = new ACLMessage(ACLMessage.PROPOSE);
						msgCFPPropose.addReceiver(restAreaManager);
						msgCFPPropose.setInReplyTo(replyCode);
						//The journey manager sends the remaining time from the new rest area to the maximun driving time
						float timeUntilMaxDriving = (newRestAreaDistanceList.get(bestIndex))/drivingSpeed * 60; //Time to arrive to the new rest area
						timeUntilMaxDriving = maximunDriveTime - (driver.getDrivingTime() + timeUntilMaxDriving);
						msgCFPPropose.setContent(String.valueOf(timeUntilMaxDriving));
						myAgent.send(msgCFPPropose);
						step++;
						break;

					}
					else {
						newRestAreaDistanceList.remove(bestIndex);
						newRestAreaList.remove(bestIndex);
						step--;
						break;

					}
				} else block();

				break;
			case 4:
				//Wait for acceptation or rejection
				mtCFPAccept = MessageTemplate.MatchInReplyTo(replyCode);

				ACLMessage msgCFPResponse = myAgent.receive(mtCFPAccept);
				
				if (msgCFPResponse != null) {

					//If the proposal is accepted, the new reservation will start
					if (msgCFPResponse.getPerformative()==ACLMessage.ACCEPT_PROPOSAL) {
						ACLMessage msgRequestParkingSpot = new ACLMessage(ACLMessage.REQUEST);
						msgRequestParkingSpot.setLanguage(codec.getName());
						msgRequestParkingSpot.setOntology(ontoloy.getName());
						msgRequestParkingSpot.addReceiver(newRestAreaList.get(bestIndex));
						reservationCode = UUID.randomUUID().toString();

						reservationRequest = new ReservationRequest();
						reservationRequest.setReservation(reservation);
						//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
						action = new Action(getAID(),reservationRequest);

						try {
							getContentManager().fillContent(msgRequestParkingSpot, action);
						} catch (CodecException | OntologyException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}

						msgRequestParkingSpot.setInReplyTo(reservationCode);
						msgRequestParkingSpot.setConversationId("request");
						myAgent.send(msgRequestParkingSpot);
						step++;
						break;
					}

					//If our proposal is rejected, the behavior is over
					if (msgCFPResponse.getPerformative()==ACLMessage.REJECT_PROPOSAL){

						step = 6;
						break;
					}
				} else block();
				break;
			case 5:
				//Send a msg to restAreaManager confirming that the reservation has been done
				MessageTemplate mtReservationReply = MessageTemplate.MatchInReplyTo(reservationCode);

				ACLMessage msgReservationReply = myAgent.receive(mtReservationReply);
				
				if (msgReservationReply != null) {

					//Everything is OK, the agent answers the original rest area that we have done a reservation
					if (msgReservationReply.getPerformative()==ACLMessage.AGREE) {
						nextRestArea = msgReservationReply.getSender();
						ACLMessage msgInform = new ACLMessage(ACLMessage.INFORM);
						msgInform.addReceiver(restAreaManager);
						msgInform.setInReplyTo(replyCode);
						myAgent.send(msgInform);
						nextRestArea = newRestAreaList.get(bestIndex);
						estimatedArrivalTime.setTime(reservation.getTimeStart());
						estimatedArrivalTime.setTime(reservation.getTimeEnd());
						step++;
						break;
					}

					//If there is a problem, the agent answers the original rest area that we have not done a reservation
					if (msgReservationReply.getPerformative()==ACLMessage.REFUSE) {
						ACLMessage msgInform = new ACLMessage(ACLMessage.FAILURE);
						msgInform.addReceiver(restAreaManager);
						msgInform.setInReplyTo(replyCode);
						myAgent.send(msgInform);
						step++;
						break;
					}
				} else block();
				break;
			}
		}

		@Override
		public boolean done() {
			if (step == 6) {
				addBehaviour(new BJourneyManagerNegotiation());
				return true;
			}
			// TODO Auto-generated method stub
			return false;
		}
	}

	private class BJourneyManagerUpdate extends Behaviour {
		int minutes = 0;
		int km_aux = 0;

		@Override
		public void action() {
			// TODO Auto-generated method stub
			MessageTemplate mtInformUpdate = MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.INFORM), MessageTemplate.MatchConversationId("timeUpdate"));

			ACLMessage msgInformUpdate = myAgent.receive(mtInformUpdate);
			if (msgInformUpdate != null) {
				try {
					ContentElement content = getContentManager().extractContent(msgInformUpdate);
					UpdateTime updateTime = (UpdateTime) ((Action) content).getAction();
					actualTime.setTime(updateTime.getTimeUpdated());
					minutes = updateTime.getMinutesUpdated();
					km_aux = Math.round(drivingSpeed * minutes / 60);
					//The truck is in the rest area
					if ((actualTime.before(estimatedEndingTime) || actualTime.equals(estimatedEndingTime)) && (actualTime.after(estimatedArrivalTime) || actualTime.equals(estimatedArrivalTime))) {
						resting = true;
						if (drivedTimeBeforeRest != 0)
							lostTime = lostTime + (maximunDriveTime - drivedTimeBeforeRest);
						drivedTimeBeforeRest = 0;
					}
					//The truck advances
					else {
						drivedTimeBeforeRest = drivedTimeBeforeRest + minutes;
						drivedTime = drivedTime + minutes;
						km = km + km_aux;
						//If the truck reaches certain point of the road, it creates a log and disapear from the system
						if (km >= km_out) {
							myAgent.doDelete();
						}
						//If the journey manager leaves the rest area, it must do another reservation
						if (resting) {
							long restTime = (estimatedEndingTime.getTimeInMillis() - estimatedArrivalTime.getTimeInMillis())/60000; //from ms to min
							restedTime = restedTime + (int)restTime;
							resting = false;
							myAgent.addBehaviour(new BJourneyManagerReservation());
						}
					}
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} else block();

		}

		@Override
		public boolean done() {
			// TODO Auto-generated method stub
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//

	
	//******************************************************************************************//
	//******************************************************************************************//
	// Simulation without negotiation behaviours: BJourneyManagerReservationFree,				//
	//											  BJourneyManagerUpdateFree						//
	//******************************************************************************************//
	//******************************************************************************************//
	private class BJourneyManagerReservationFree extends Behaviour {

		int step = 0;
		int bestIndex = 0;
		int index = 0;
		float bestDistance;
		String reservationCode;

		@Override
		public void action() {
			switch (step){
			//Creating the Preference List
			case 0:
				if (!creatingList)	{
					addBehaviour(new BJourneyManagerCreatingPreferencesTable());
					creatingList = true;
				}
				if (listCompleted) {
					step++;
					creatingList = false;
				}
				break;
			//Choosing the best rest area to park (it is the rest area that allows the truck drive the maximun distance
			case 1:
				//If the journey manager runs out of options, it must exit of the system
				if (RestAreaList.isEmpty()){
					infractionCommited = 1;
					myAgent.doDelete();
					step = 5;
					break;
				}
				bestIndex = 0;
				index = 0;
				bestDistance = 0;
				//The best will be the one most far away from my position (the one which is value is greater)
				for (float distance : RestAreaDistanceList){
					if (distance > bestDistance) {
						bestIndex = index;
						bestDistance = distance;
					}
					index++;
				}
				//This will be the target rest area
				nextRestArea = RestAreaList.get(bestIndex);
				Calendar timeStart = Calendar.getInstance(), timeEnd = Calendar.getInstance();
				timeStart.setTime(actualTime.getTime());
				timeEnd.setTime(actualTime.getTime());
				timeStart.add(Calendar.MINUTE,(int) (bestDistance*60/drivingSpeed));
				timeEnd.add(Calendar.MINUTE,(int) (bestDistance*60/drivingSpeed)+restTime);
				estimatedArrivalTime.setTime(timeStart.getTime());
				estimatedEndingTime.setTime(timeEnd.getTime());
				//
				step++;
				break;
			case 2:
				if (reachRestArea) {
					reservationCode = UUID.randomUUID().toString();
					ACLMessage msgRequestParkingSpot = new ACLMessage(ACLMessage.REQUEST);
					msgRequestParkingSpot.addReceiver(nextRestArea);
					msgRequestParkingSpot.setConversationId("requestFree");
					msgRequestParkingSpot.setInReplyTo(reservationCode);
					//Send the reservation
					myAgent.send(msgRequestParkingSpot);
					step++;
					break;
				}
				break;
			case 3:
				MessageTemplate mtReservationReply = MessageTemplate.MatchInReplyTo(reservationCode);

				ACLMessage msgReservationReply = myAgent.receive(mtReservationReply);

				if (msgReservationReply != null) {
					//The journey manager dida reservation
					if (msgReservationReply.getPerformative() == ACLMessage.AGREE){
						resting = true;
						step++;
						break;
					}
					//The journey manager couldn't rest in the desired area
					else {
						infractionCommited = 1;
						myAgent.doDelete();
						step = 5;
						break;
					}
				}
				break;
			case 4:
				if (actualTime.after(estimatedEndingTime)) {
					//The Journey Manager requests to the Rest Area Manager to free the parking spot
					ACLMessage msgRequestFreeParkingSpot = new ACLMessage(ACLMessage.REQUEST);
					msgRequestFreeParkingSpot.addReceiver(nextRestArea);
					msgRequestFreeParkingSpot.setConversationId("requestSetSpotFree");
					//Send the reservation
					myAgent.send(msgRequestFreeParkingSpot);
					step++;
					break;
				}
				break;
			}
		}

		@Override
		public boolean done() {
			if (step == 5) return true;
			return false;
		}
	}
	private class BJourneyManagerUpdateFree extends Behaviour {
		int minutes = 0;
		int km_aux = 0;

		@Override
		public void action() {
			// TODO Auto-generated method stub
			MessageTemplate mtInformUpdate = MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.INFORM), MessageTemplate.MatchConversationId("timeUpdate"));

			ACLMessage msgInformUpdate = myAgent.receive(mtInformUpdate);
			if (msgInformUpdate != null) {
				try {
					ContentElement content = getContentManager().extractContent(msgInformUpdate);
					UpdateTime updateTime = (UpdateTime) ((Action) content).getAction();
					actualTime.setTime(updateTime.getTimeUpdated());
					minutes = updateTime.getMinutesUpdated();
					km_aux = Math.round(drivingSpeed * minutes / 60);
					//The truck is in the rest area
					if ((actualTime.before(estimatedEndingTime) || actualTime.equals(estimatedEndingTime)) && (actualTime.after(estimatedArrivalTime) || actualTime.equals(estimatedArrivalTime))) {
						reachRestArea = true;
						if (drivedTimeBeforeRest != 0)
							lostTime = lostTime + (maximunDriveTime - drivedTimeBeforeRest);
						drivedTimeBeforeRest = 0;
					}
					//The truck advances
					else {
						drivedTimeBeforeRest = drivedTimeBeforeRest + minutes;
						drivedTime = drivedTime + minutes;
						km = km + km_aux;
						//If the truck reaches certain point of the road, it creates a log and disapear from the system
						if (km >= km_out) {
							myAgent.doDelete();
						}
						if (resting) {
							reachRestArea = false;
							resting = false;
							long restTime = (estimatedEndingTime.getTimeInMillis() - estimatedArrivalTime.getTimeInMillis())/60000; //from ms to min
							restedTime = restedTime + (int)restTime;
							myAgent.addBehaviour(new BJourneyManagerReservationFree());
						}
					}
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} else block();

		}

		@Override
		public boolean done() {
			// TODO Auto-generated method stub
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//

	
	//******************************************************************************************//
	//******************************************************************************************//
	// General behavious: BJourneyManagerCreatingPreferencesTable, BJourneyManagerShutDown		//
	//******************************************************************************************//
	//******************************************************************************************//	
	private class BJourneyManagerCreatingPreferencesTable extends Behaviour {
		int step = 0;
		DFAgentDescription dfd = null;
		DFAgentDescription[] result = null;
		int numParkingSpots = 0;

		@Override
		public void action() {
			// TODO Auto-generated method stub

			switch(step) {
			case 0:	
				//Send a message to all Rest Area agents in the same road of the truck to ask AIDs
				ServiceDescription sd = new ServiceDescription(); 
				sd.setType("RestArea");
				sd.setName(road.getName());
				dfd = new DFAgentDescription();
				dfd.addServices(sd);
				step++;
				break;
			case 1: 
				//Search for 0.1 sec
				while (result == null) {
				    try {
					  result = DFService.searchUntilFound(
							  myAgent, getDefaultDF(), dfd, null, 100);
					} catch (FIPAException e) { e.printStackTrace();}
				}
			    if (result.length >= 1) step++;
			    break;
			case 2:
				String localName;
				float distance;

				for (DFAgentDescription restArea : result) {
					localName = restArea.getName().getLocalName();
					distance = Float.parseFloat(localName.substring(localName.indexOf(road.getName())+road.getName().length()));
					//The only rest areas that could be of use, are those which are at not to far away or behind of the truck
					if ((distance  > km) && (distance < km + ((maximunDriveTime - (driver.getDrivingTime())*drivingSpeed/60)))) {
						RestAreaList.add(restArea.getName());
						RestAreaDistanceList.add((distance-km));
						RestAreaKMList.add(distance);
					}
				}
				step++;
				break;
			}
			}

		@Override
		public boolean done() {
			if (step == 3) {
				listCompleted = true;
				return true;
			}
			return false;
		}
	}

	private class BJourneyManagerShutDown extends Behaviour {

		@Override
		public void action() {
			// TODO Auto-generated method stub
			MessageTemplate mtInformShutDown = MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.INFORM), MessageTemplate.MatchConversationId("shutDown"));

			ACLMessage msgInformUpdate = myAgent.receive(mtInformShutDown);
			if (msgInformUpdate != null) {
					myAgent.doDelete();
				}
			}

		@Override
		public boolean done() {
			// TODO Auto-generated method stub
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//	

}
