package agents;

import gui.LoaderGUI;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.CyclicBehaviour;
import jade.core.behaviours.OneShotBehaviour;
import jade.core.behaviours.TickerBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.domain.JADEAgentManagement.JADEManagementOntology;
import jade.domain.JADEAgentManagement.ShutdownPlatform;
import jade.gui.GuiAgent;
import jade.gui.GuiEvent;
import jade.wrapper.AgentController;
import jade.wrapper.ContainerController;
import jade.wrapper.StaleProxyException;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Random;
import java.util.StringTokenizer;

import ontology.Driver;
import ontology.ParkingManagementOntology;
import ontology.Reservation;
import ontology.ReservationRequest;
import ontology.RestArea;
import ontology.Road;
import ontology.UpdateTime;
import ontology.Vehicle;


/**
 * @author Raquel Garca Arvalo
 * @code: Loader agent code
 * 
 */

public class AgLoader extends GuiAgent {
	private Codec codec = new SLCodec();
	private Ontology ontologia = ParkingManagementOntology.getInstance();
	private AID ID;
	private gui.GUIData SetupData;
	private int createdVehicles = 0;
	private AgentController ac;
	private Calendar SimulationStart = Calendar.getInstance();
	private Calendar SimulatedNow = Calendar.getInstance();
	private LoaderGUI myGUI;
	private ArrayList<Road> roadList = new ArrayList<Road>();
	private PrintWriter wrVehicleFile;

	@Override
	protected void setup() {
		// Register codec and ontology
		getContentManager().registerLanguage(codec);
		getContentManager().registerOntology(ontologia);

		// Register the offered service
		DFAgentDescription dfd = new DFAgentDescription();
		ID = getAID();
		dfd.setName(ID);
		ServiceDescription sd = new ServiceDescription();
		sd.setType("Loader");
		sd.setName("loader");
		dfd.addServices(sd);
		try {
			DFService.register(this, dfd);
		} catch (Exception e) {
			e.printStackTrace();
		}
		

		// Open the GUI
		myGUI = new gui.LoaderGUI(this);
		myGUI.setVisible(true);

		//Set time
		SimulationStart.setTime(Calendar.getInstance().getTime());
		SimulatedNow.setTime(Calendar.getInstance().getTime());
	}
	
	@Override
	protected void takeDown()
	{
		//Warn  the agents of the system of the shut down of the system
		DFAgentDescription[] resultJM = null;
		ACLMessage msgShutDownJM = new ACLMessage(ACLMessage.INFORM);
		msgShutDownJM.setConversationId("shutDown");

		//Add all the journey managers
		ServiceDescription sdJourneyManager = new ServiceDescription();
		sdJourneyManager.setType("JourneyManager");
		DFAgentDescription dfdJourneyManager = new DFAgentDescription();
		dfdJourneyManager.addServices(sdJourneyManager);
		
		while (resultJM == null) {
		    try {
		    	resultJM = DFService.searchUntilFound(this, getDefaultDF(), dfdJourneyManager, null, 500);
			} catch (FIPAException e) { 
				e.printStackTrace(); 
				System.out.println("FIPA Error while searching for agents");
			}
		}

	    if (resultJM.length > 0) 
	    	for (DFAgentDescription agent : resultJM) {
	    		msgShutDownJM.addReceiver(agent.getName());
	    	}
	    
	    //Send the shut down msg and wait 0.5 sec to allow journey managers to send their information
	    send(msgShutDownJM);
	    
	    System.out.println("Closing journey managers, this could take some time...");
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		//Warn  the agents of the system of the shut down of the system
		DFAgentDescription[] resultRA = null;
		ACLMessage msgShutDownRA = new ACLMessage(ACLMessage.INFORM);
		msgShutDownRA.setConversationId("shutDown");

		//Add all the rest areas
		ServiceDescription sdRestAreaManager = new ServiceDescription();
		sdRestAreaManager.setType("RestArea");
		DFAgentDescription dfdRestAreaManager = new DFAgentDescription();
		dfdRestAreaManager.addServices(sdRestAreaManager);
		
		while (resultRA == null) {
		    try {
		    	resultRA = DFService.searchUntilFound(this, getDefaultDF(), dfdRestAreaManager, null, 500);
			} catch (FIPAException e) { 
				e.printStackTrace(); 
				System.out.println("FIPA Error while searching for agents");
			}
		}

	    if (resultRA.length > 0)
	    	for (DFAgentDescription agent : resultRA) {
	    		msgShutDownRA.addReceiver(agent.getName());
	    	}
	    
	    //Send the shut down msg and wait 0.5 sec to allow journey managers to send their information
	    send(msgShutDownRA);
	    
	    System.out.println("Closing rest area managers, this could take some time...");
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		//Warn  the agents of the system of the shut down of the system
		DFAgentDescription[] resultPS = null;
		ACLMessage msgShutDownPS = new ACLMessage(ACLMessage.INFORM);
		msgShutDownPS.setConversationId("shutDown");

		//Add all the parking spots
		ServiceDescription sdParkingSpots = new ServiceDescription();
		sdRestAreaManager.setType("ParkingSpot");
		DFAgentDescription dfdParkingSpots = new DFAgentDescription();
		dfdParkingSpots.addServices(sdParkingSpots);
		
		while (resultPS == null) {
		    try {
		    	resultPS = DFService.searchUntilFound(this, getDefaultDF(), dfdParkingSpots, null, 500);
			} catch (FIPAException e) { 
				e.printStackTrace(); 
				System.out.println("FIPA Error while searching for agents");
			}
		}

	    if (resultPS.length > 0)
	    	for (DFAgentDescription agent : resultPS) {
	    		msgShutDownPS.addReceiver(agent.getName());
	    	}
	    
	    send(msgShutDownPS);

	    System.out.println("Closing parking spot managers, this could take some time...");
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		//Warn  the agents of the system of the shut down of the system
		DFAgentDescription[] resultLog = null;
		ACLMessage msgShutDownLog = new ACLMessage(ACLMessage.INFORM);
		msgShutDownLog.setConversationId("shutDown");

		//Add the log agent
		ServiceDescription sdLogManager = new ServiceDescription();
		sdLogManager.setType("LogManager");
		sdLogManager.setName("logManager");
		DFAgentDescription dfdLogManager = new DFAgentDescription();
		dfdLogManager.addServices(sdLogManager);
		
		while (resultLog == null) {
		    try {
		    	resultLog = DFService.searchUntilFound(this, getDefaultDF(), dfdLogManager, null, 500);
			} catch (FIPAException e) { 
				e.printStackTrace(); 
				System.out.println("FIPA Error while searching for agents");
			}
		}

	    if (resultLog.length > 0)
	    	for (DFAgentDescription agent : resultLog) {
	    		msgShutDownLog.addReceiver(agent.getName());
	    	}

		//Send the message
		send(msgShutDownLog);
		
	    System.out.println("Closing log manager, this could take some time...");
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		System.out.println("Closing...");
		//If we didn't start the simulation, SetupData is null.
		if (SetupData != null && SetupData.i_CreateLog) {
			wrVehicleFile.close();
		}
		
		if (myGUI != null)
		{
			myGUI.setVisible(false);
			
			//Asking the AMS to shut down the platform
			Codec codec = new SLCodec();    
			Ontology jmo = JADEManagementOntology.getInstance();
			getContentManager().registerLanguage(codec);
			getContentManager().registerOntology(jmo);
			ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
			msg.addReceiver(getAMS());
			msg.setLanguage(codec.getName());
			msg.setOntology(jmo.getName());
			try {
			    getContentManager().fillContent(msg, new Action(getAID(), new ShutdownPlatform()));
			    send(msg);
			}
			catch (Exception e) {}
		}
	}

	Road findRoad(int id) {
		for (Road road : roadList) {
			if (road.getRoadId() == id) {
				return road;
			}
		}
		return null;
	}

	@Override
	protected void onGuiEvent(GuiEvent ev) {

		int command = ev.getType();

		if (command == 0) 
		{
			this.doDelete();
		} 
		else if (command == 1) {
			SetupData = (gui.GUIData) ev.getParameter(0);

			File FilesPath;
			FilesPath = new File(SetupData.i_setupFilesFolder);

			ContainerController cc = getContainerController();
			
			try {
				Object[] args = new Object[1];
				args[0] = FilesPath.toString();
				
				ac = cc.createNewAgent("LogManager",
						"agents.AgLogManager", args);
				ac.start();
			} catch (StaleProxyException e) {
				System.out.println("Error creating LogManager Agent");
				e.printStackTrace();
			}			
			
			Charset charset = Charset.forName("US-ASCII");

			//The created agents need some time to finish their setup
			System.out.println("Creating log manager, this could take some seconds...");
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			//

			// ***************** Road *********************************
			Path RoadsFile = Paths.get(FilesPath.getAbsolutePath() + "\\Roads");
			try (BufferedReader reader1 = Files.newBufferedReader(RoadsFile,
					charset)) {
				String line = null;

				while ((line = reader1.readLine()) != null) {

					StringTokenizer fields = new StringTokenizer(line, ",");

					Road r = new Road();
					r.setRoadId(Integer.parseInt(fields.nextToken().trim()));
					r.setName(fields.nextToken().trim());
					r.setLenght(Integer.parseInt(fields.nextToken().trim()));
					r.setTraffic(Float.parseFloat(fields.nextToken().trim()));
					roadList.add(r);

				}
			} catch (IOException x) {
				System.out
						.println("Couldn't read file:" + RoadsFile.toString());
				System.err.format("IOException: %s%n", x);
			}


			// ***************** REST AREAS CREATION *************************
			// Search for the RestAreas file
			Path RestAreasFile = Paths.get(FilesPath.getAbsolutePath()
					+ "\\RestAreas");
			try (BufferedReader reader2 = Files.newBufferedReader(
					RestAreasFile, charset)) {
				String line = null;
				while ((line = reader2.readLine()) != null) {
					StringTokenizer fields = new StringTokenizer(line, ",");

					RestArea ra = new RestArea();

					ra.setRestAreaId(Integer.parseInt(fields.nextToken().trim()));

					int roadID = Integer.parseInt(fields.nextToken().trim());

					try {
						ra.setRoad(findRoad(roadID));
					} catch (NullPointerException e) {
						System.out.println("Road for the Rest Area not found");
						e.printStackTrace();
					}

					ra.setKM(Integer.parseInt(fields.nextToken().trim()));
					ra.setName(fields.nextToken().trim());
					ra.setAreaSpots(Integer.parseInt(fields.nextToken().trim()));

					// Create agent
					try {
						Object[] args = new Object[2];
						args[0] = ra;
						args[1] = SetupData.i_simulationType;

						ac = cc.createNewAgent("" + ra.getName() + ""
								+ ra.getRoad().getName() + ra.getKM(),
								"agents.AgRestAreaManager", args);
						ac.start();
					} catch (StaleProxyException e) {
						System.out.println("Error creating AgRestAreaManager");
						e.printStackTrace();
					}

				}
			} catch (IOException x) {
				System.out.println("Couldn't read file:"
						+ RestAreasFile.toString());
				System.err.format("IOException: %s%n", x);
			}


			//The created agents need some time to finish their setup
			System.out.println("Creating rest areas, this could take some seconds...");
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			//

			if (SetupData.i_createRandomVehicles) {

				if (SetupData.i_CreateLog) {
					try {
						wrVehicleFile = new PrintWriter(
								SetupData.i_setupFilesFolder.toString() + "\\"
										+ SetupData.i_vehicleLogName, "UTF-8");
					} catch (FileNotFoundException
							| UnsupportedEncodingException e) {

						e.printStackTrace();
					}

				}


				addBehaviour(new CreateJourneyManagerAgent(this, SetupData.i_vehicleCreationLapse * 1000));
				
			} else {
				// ***************** JOURNEY MANAGERS CREATION
				// *************************
				// Search for the Vehicles file
				Path VehiclesFile = Paths.get(FilesPath.getAbsolutePath()
						+ "\\Vehicles");
				try (BufferedReader reader3 = Files.newBufferedReader(
						VehiclesFile, charset)) {
					String line = null;
					while ((line = reader3.readLine()) != null) {
						StringTokenizer fields = new StringTokenizer(line, ",");

						Vehicle v = new Vehicle();

						v.setVehicleId(Integer.parseInt(fields.nextToken()
								.trim()));
						v.setMaxSpeed(Integer.parseInt(fields.nextToken()
								.trim()));

						int roadID = Integer
								.parseInt(fields.nextToken().trim());
						v.setRoad(findRoad(roadID));

						v.setKM(Integer.parseInt(fields.nextToken().trim()));

						Driver vehicleDriver = new Driver();
						vehicleDriver.setDriverId(Integer.parseInt(fields.nextToken().trim()));
						vehicleDriver.setName(fields.nextToken().trim());
						vehicleDriver.setDrivingTime(Integer.parseInt(fields
								.nextToken().trim()));
						v.setDriver(vehicleDriver);

						// Create agent
						try {
							Object[] args = new Object[3];
							args[0] = v;
							args[1] = SetupData.i_simulationType;
							args[2] = SimulatedNow;
							
							ac = cc.createNewAgent("JM" + v.getVehicleId(),
									"agents.AgJourneyManager", args);
							ac.start();
						} catch (StaleProxyException e) {
							e.printStackTrace();
						}

					}
				} catch (IOException x) {
					System.err.format("IOException: %s%n", x);
				}
				// END OF AGENTS CREATION
			}

			//The created agents need some time to finish their setup
			System.out.println("Creating journey managers, this could take some seconds...");
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			//

			addBehaviour(new BUpdateTime(this, SetupData.i_simulationUpdate));
		}

	}

	private class CreateJourneyManager extends OneShotBehaviour {


		private Vehicle CreateRandomVehicle() {
			Random myRandomizer = new Random();
			createdVehicles++;
			Vehicle v = new Vehicle();
			Driver d = new Driver();
			d.setDriverId(createdVehicles);
			d.setName("Name" + createdVehicles);
			d.setDrivingTime(myRandomizer.nextInt(20) * 10); //Random drived time from 0 up to 200 minutes
			v.setDriver(d);
			v.setVehicleId(createdVehicles);
			v.setMaxSpeed((myRandomizer.nextInt(4) + 6) * 10); //Random speed from 60 up to 100 km/h
			v.setRoad(roadList.get(myRandomizer.nextInt(roadList.size())));
			v.setKM(myRandomizer.nextInt(v.getRoad().getLenght()));
			return v;
		}

		@Override
		public void action() {
			try {
				Vehicle rv = CreateRandomVehicle();
				Object[] args = new Object[3];
				args[0] = rv;
				args[1] = SetupData.i_simulationType;
				args[2] = SimulatedNow;
				
				ContainerController cc = getContainerController();
				ac = cc.createNewAgent("JM" + rv.getVehicleId(),
						"agents.AgJourneyManager", args);

				if (SetupData.i_CreateLog) {
					String vdata = rv.getVehicleId() + "," + rv.getMaxSpeed() + ","
							+ rv.getRoad().getRoadId() + "," + rv.getKM() + ","
							+ rv.getDriver().getDriverId() + ","
							+ rv.getDriver().getName() + ","
							+ rv.getDriver().getDrivingTime();
					wrVehicleFile.println(vdata);
				}

				ac.start();
			} catch (StaleProxyException e) {
				e.printStackTrace();
			}

		}

	}
	
	private class CreateJourneyManagerAgent extends TickerBehaviour {

		public CreateJourneyManagerAgent(Agent a, long period) {
			super(a, period);
		}

		@Override
		protected void onTick() {
			for (int i = 0; i < SetupData.i_numberOfVehicles; i++) {
				addBehaviour(new CreateJourneyManager());
			}			
		}
		
	}
	
	private class BUpdateTime extends TickerBehaviour {
		//Ontology
		private Codec codec = new SLCodec();
		private Ontology ontoloy = ParkingManagementOntology.getInstance();
		Action action;

		public BUpdateTime(Agent a, long period) {
			super(a, period);
			// TODO Auto-generated constructor stub
		}

		@Override
		protected void onTick() {
			DFAgentDescription[] result = null;
			ACLMessage msgTimeUpdate = new ACLMessage(ACLMessage.INFORM);
			msgTimeUpdate.setConversationId("timeUpdate");

			//Hacer una bsqueda de todos los journeyManager que estn "vivos"
			ServiceDescription sd = new ServiceDescription(); 
			sd.setType("JourneyManager");
			DFAgentDescription dfd = new DFAgentDescription();
			dfd.addServices(sd);

			//Search for a tenth of the waiting time of the tick behaviour 
			    try {
				  result = DFService.searchUntilFound(myAgent, getDefaultDF(), dfd, null, getPeriod()/10);
				} catch (FIPAException e) { 
					e.printStackTrace(); 
					System.out.println("FIPA Error while searching for agents");
				}

			//Update time adding the lapse
			SimulatedNow.add(Calendar.MINUTE, SetupData.i_simMinutesPerLapse);

			//Creating the reservation request message using ontology
			msgTimeUpdate.setLanguage(codec.getName());
			msgTimeUpdate.setOntology(ontoloy.getName());
			UpdateTime updateTime = new UpdateTime();
			updateTime.setTimeUpdated(SimulatedNow.getTime());
			updateTime.setMinutesUpdated(SetupData.i_simMinutesPerLapse);
			//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
			action = new Action(getAID(),updateTime);
			try {
				getContentManager().fillContent(msgTimeUpdate, action);
			} catch (CodecException | OntologyException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			long diference = (SimulatedNow.getTimeInMillis() - SimulationStart.getTimeInMillis()) /60000;

			myGUI.setStatusProgress((int)diference,SimulatedNow.getTime());
			
			if (diference < SetupData.i_simulationTime) {
				//msgTimeUpdate.setContent(SimulatedNow.getTime().toString());
					
				//If we found at least one Journey Manager, we send them the new time
				if (result != null && result.length >= 1) {	
					//Add all journey manager to the msg
					for (DFAgentDescription journeyManager : result) {
						msgTimeUpdate.addReceiver(journeyManager.getName());
					}
					//Send the update time msg
					myAgent.send(msgTimeUpdate);
				}
			}
			else {
				//If we have reached the end time, we close.
				doDelete();
			}
		}
		
	}
}

