package agents;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;

import jade.content.Concept;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.CyclicBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import ontology.LogJourneyData;
import ontology.LogRelocation;
import ontology.ParkingManagementOntology;

/**
 * @author Raquel Garca Arvalo
 * @code: Log Manager Agent code
 * 
 */

public class AgLogManager extends Agent {
	private Codec codec = new SLCodec();
	private Ontology ontologia = ParkingManagementOntology.getInstance();
	private AID ID;
	
	private String filesPath;
	private PrintWriter wrInfractionsLog;
	private PrintWriter wrRelocationsLog;
	private PrintWriter wrJourneyDataLog;

	@Override
	protected void setup() {
		// Register codec and ontology
		getContentManager().registerLanguage(codec);
		getContentManager().registerOntology(ontologia);

		// Register the offered service
		DFAgentDescription dfd = new DFAgentDescription();
		ID = getAID();
		dfd.setName(ID);
		ServiceDescription sd = new ServiceDescription();
		sd.setType("LogManager");
		sd.setName("logManager");
		dfd.addServices(sd);
		try {
			DFService.register(this, dfd);
		} catch (Exception e) {
			e.printStackTrace();
		}

		//Load files path
		Object[] args = getArguments();
		filesPath = (String) args[0];
		
		
		
		//Create files
		try {
			wrJourneyDataLog = new PrintWriter(
					filesPath + "\\JourneyDataLog.txt", "UTF-8");
		} catch (FileNotFoundException
				| UnsupportedEncodingException e) {

			e.printStackTrace();
		}
		
		try {
			wrRelocationsLog = new PrintWriter(
					filesPath + "\\RelocationsLog.txt", "UTF-8");
		} catch (FileNotFoundException
				| UnsupportedEncodingException e) {

			e.printStackTrace();
		}
		
		
		addBehaviour(new BLogMessageReceiver());
		addBehaviour(new BLogManagerShutDown());
		
	}

	
	@Override
	protected void takeDown()
	{
		//Close files
		wrRelocationsLog.close();
		wrJourneyDataLog.close();
		
	}
	
	private class BLogMessageReceiver extends CyclicBehaviour {

		@Override
		public void action() {
			ACLMessage msg = receive(MessageTemplate.MatchConversationId("log"));
			if (msg == null) {
				block();
				return;
			}

			try {
				ContentElement content = getContentManager()
						.extractContent(msg);
				Concept action = ((Action) content).getAction();
				switch (msg.getPerformative()) {
				case (ACLMessage.INFORM):

					if (action instanceof LogJourneyData)
					{
						String journeyDataLine = "";
						LogJourneyData j = (LogJourneyData) action;
						journeyDataLine = j.getIdVehicle()+ ",";
						journeyDataLine += j.getdrivedTime()+ ",";
						journeyDataLine += j.getKm()+ ",";
						journeyDataLine += j.getKmIn()+ ",";
						journeyDataLine += j.getKmOut()+ ",";
						journeyDataLine += j.getTimeIn()+ ",";
						journeyDataLine += j.getTimeOut()+ ",";
						journeyDataLine += j.getlostTime()+ ",";
						journeyDataLine += j.getrestTime()+ ",";
						journeyDataLine += j.getInfractionType();
						
						wrJourneyDataLog.println(journeyDataLine);
						
						
					}
					
					if (action instanceof LogRelocation)
					{
						
						String relocationLine = "";
						LogRelocation r = (LogRelocation) action;
						relocationLine = r.getRelocationTime().toString()+ ",";
						relocationLine += r.getAffectedVehicles()+ ",";
						relocationLine += r.getRelocableVehicles()+ ",";
						relocationLine += r.getRelocatedVehicleOrder()+ ",";
						relocationLine += r.getSuccess();
						
						wrRelocationsLog.println(relocationLine);
						
					}
					break;

				}
			} catch (Exception ex) {
				ex.printStackTrace();
			}
		}
	}
	
	private class BLogManagerShutDown extends Behaviour {

		@Override
		public void action() {
			// TODO Auto-generated method stub
			MessageTemplate mtInformShutDown = MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.INFORM), MessageTemplate.MatchConversationId("shutDown"));

			ACLMessage msgInformUpdate = myAgent.receive(mtInformShutDown);
			if (msgInformUpdate != null) {
					myAgent.doDelete();
				}
			}

		@Override
		public boolean done() {
			// TODO Auto-generated method stub
			return false;
		}
	}

}

