package agents;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Random;
import java.util.UUID;

import ontology.CancelRequest;
import ontology.OccupationRequest;
import ontology.ParkingManagementOntology;
import ontology.Reservation;
import ontology.ReservationRequest;
import jade.content.Concept;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.UngroundedException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.CyclicBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;


/**
 * @author Francisco Palau Romero
 * @code: Parking Spot agent code 
 *
 */

public class AgParkingSpot extends Agent {
	private AID ID;
	private class reservationList {
		Calendar day;
		ArrayList<Reservation> list = new ArrayList<Reservation>(); 
	}
	private reservationList todayReservations = new reservationList();
	private reservationList tomorrowReservations = new reservationList();
	//Ontology
	private Codec codec = new SLCodec();
	private Ontology ontoloy = ParkingManagementOntology.getInstance();
	
	
	//******************************************************************************************//
	//******************************************************************************************//	
	// Setup																					//
	//******************************************************************************************//
	//******************************************************************************************//
	protected void setup() {
		//Register codec and ontology
	    getContentManager().registerLanguage(codec);
	    getContentManager().registerOntology(ontoloy);

	    //Register the offered service
		DFAgentDescription dfd = new DFAgentDescription();
		ID = getAID();
		dfd.setName(ID);
		ServiceDescription sd = new ServiceDescription();
		sd.setType("ParkingSpot");
		sd.setName("Spot");
		dfd.addServices(sd);
		try{
			DFService.register(this, dfd);
		} catch (Exception e) {e.printStackTrace();}


		//When it is registered, the rest area agent begins his behaviour
		addBehaviour(new BParkingSpotMessageReceiver());
		addBehaviour(new BParkingSpotShutDown());
	}
	//******************************************************************************************//
	//******************************************************************************************//	

	
	//******************************************************************************************//
	//******************************************************************************************//	
	// Override																					//
	//******************************************************************************************//
	//******************************************************************************************//
	@Override
	protected void takeDown() {
		while (receive() != null) {};
	}
	//******************************************************************************************//
	//******************************************************************************************//	

	
	//******************************************************************************************//
	//******************************************************************************************//	
	// Reservation behaviours: 	BParkingSpotMessageReceiver, BParkingSpotManagerCancelRequest	//
	//							BParkingSpotManagerReservationRequest							//
	//							BParkingSpotManagerOccupationRequest							//
	//******************************************************************************************//
	//******************************************************************************************//	
	private class BParkingSpotMessageReceiver extends CyclicBehaviour {
		public void action(){
			ACLMessage msg = receive(MessageTemplate.MatchConversationId("request"));
			if (msg==null){
				block();
				return;
			}

	       try {
		        ContentElement content = getContentManager().extractContent(msg);
		        Concept action = ((Action) content).getAction();
		        switch (msg.getPerformative()) { 
	         		case (ACLMessage.REQUEST):	
	         			if (action instanceof OccupationRequest) {
	         				addBehaviour(new BParkingSpotManagerOccupationRequest(msg));
	         			}
	         			else if (action instanceof CancelRequest){
	         				addBehaviour(new BParkingSpotManagerCancelRequest(msg));
	         			}
	         			break;
	         		case (ACLMessage.CFP):
	         			if (action instanceof ReservationRequest) {
	         				addBehaviour(new BParkingSpotManagerReservationRequest(msg));
	         			}
	         			break;
		         }
	       }
	       catch(Exception ex) { ex.printStackTrace(); }
		}
	}
	

	private class BParkingSpotManagerReservationRequest extends Behaviour {

		int step = 0;
		ACLMessage msg;
		Calendar timeStart, timeEnd;
		Calendar today = Calendar.getInstance();;
		String journeyManagerId;
		AID restArea;
		String reservationCode;
		Boolean startToday;
		int index = 0;
		Action action;

		public BParkingSpotManagerReservationRequest(ACLMessage msg) {
			// TODO Auto-generated constructor stub
			this.msg = msg;
		}

		@Override
		public void action() {
			switch (step) {
			case 0:
				//Reading the message
				restArea = msg.getSender();
				reservationCode = msg.getInReplyTo();
				try {
					ContentElement content = getContentManager().extractContent(msg);
					ReservationRequest reservationRequest = (ReservationRequest) ((Action) content).getAction();
					Reservation reservation = reservationRequest.getReservation();
					timeStart = Calendar.getInstance();
					timeEnd = Calendar.getInstance();
					timeStart.setTime(reservation.getTimeStart());
					timeEnd.setTime(reservation.getTimeEnd());
					journeyManagerId = reservation.getjourneyManagerId();
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				step++;
				break;
			case 1:
				//Check for free space
				startToday = (timeStart.get(Calendar.DAY_OF_YEAR) == today.get(Calendar.DAY_OF_YEAR));
				//The reservation will start today
				if (startToday) {
					//If the list is empty...
					if (todayReservations.list.isEmpty()) {
						step++;
						break;
					}
					//Check for all reservations for this day
					for ( Reservation reservation : todayReservations.list) {
						Calendar timeStartReservation = Calendar.getInstance();
						timeStartReservation.setTime(reservation.getTimeStart());
						Calendar timeEndReservation = Calendar.getInstance();
						timeEndReservation.setTime(reservation.getTimeEnd());
						//The asked time reservation will end in an occupied block of time
						if (timeStart.before(timeStartReservation) && timeEnd.after(timeStartReservation)){
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(reservationCode);
							msgReplyRefuse.setContent(reservation.getjourneyManagerId());
							myAgent.send(msgReplyRefuse);
							step = 4;
							break;
						}
						//The asked time reservation will start between an occupied block of time
						if ((timeStart.after(timeStartReservation) && timeStart.before(timeEndReservation)) || timeStart.equals(timeStartReservation)){
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(reservationCode);
							msgReplyRefuse.setContent(reservation.getjourneyManagerId());
							myAgent.send(msgReplyRefuse);
							step = 4;
							break;
						}
					}
				}
				//The reservation will begin tomorrow
				else {
					//If the list is empty...
					if (tomorrowReservations.list.isEmpty()) {
						step++;
						break;
					}
					//Check for all reservations for today
					for ( Reservation reservation : todayReservations.list ) {
						Calendar timeStartReservation = Calendar.getInstance();
						timeStartReservation.setTime(reservation.getTimeStart());
						Calendar timeEndReservation = Calendar.getInstance();
						timeEndReservation.setTime(reservation.getTimeEnd());
						//The asked time reservation will end in an occupied block of time
						if (timeStart.before(timeStartReservation) && timeEnd.after(timeStartReservation)){
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(reservationCode);
							msgReplyRefuse.setContent(reservation.getjourneyManagerId());
							myAgent.send(msgReplyRefuse);
							step = 4;
							break;
						}
						//The asked time reservation will start between an occupied block of time
						if ((timeStart.after(timeStartReservation) && timeStart.before(timeEndReservation)) || timeStart.equals(timeStartReservation)){
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(reservationCode);
							msgReplyRefuse.setContent(reservation.getjourneyManagerId());
							myAgent.send(msgReplyRefuse);
							step = 4;
							break;
						}
					}
					//Check for all reservations for tomorrow
					for ( Reservation reservation : tomorrowReservations.list ) {
						Calendar timeStartReservation = Calendar.getInstance();
						timeStartReservation.setTime(reservation.getTimeStart());
						Calendar timeEndReservation = Calendar.getInstance();
						timeEndReservation.setTime(reservation.getTimeEnd());
						//The asked time reservation will end in an occupied block of time
						if (timeStart.before(timeStartReservation) && timeEnd.after(timeStartReservation)){
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(reservationCode);
							msgReplyRefuse.setContent(reservation.getjourneyManagerId());
							myAgent.send(msgReplyRefuse);
							step = 4;
							break;
						}
						//The asked time reservation will start between an occupied block of time
						if ((timeStart.after(timeStartReservation) && timeStart.before(timeEndReservation)) || timeStart.equals(timeStartReservation)){
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(reservationCode);
							msgReplyRefuse.setContent(reservation.getjourneyManagerId());
							myAgent.send(msgReplyRefuse);
							step = 4;
							break;
						}
					}
				}
				step++;
				break;
			case 2:
				//fraparo: Propongo mi reserva (por el momento, envo un valor igual al conjunto de horas de ocupacin que hay en el da)
				int minutes = 0;
				Calendar calendarStart  = Calendar.getInstance(), calendarEnd  = Calendar.getInstance();

				ACLMessage msgCFPPropose = new ACLMessage(ACLMessage.PROPOSE);
				if (startToday) {
					for ( Reservation reservation : todayReservations.list) {
						calendarEnd.setTime(reservation.getTimeEnd());
						calendarStart.setTime(reservation.getTimeStart());
						minutes = calendarEnd.get(Calendar.MINUTE) - calendarStart.get(Calendar.MINUTE);
					}
				}

				msgCFPPropose.addReceiver(restArea);
				msgCFPPropose.setContent(String.valueOf(minutes));
				msgCFPPropose.setInReplyTo(reservationCode);
				myAgent.send(msgCFPPropose);
				step++;
				break;
			case 3:
				//Check if our propose is accepted
				MessageTemplate mtCFPReply = MessageTemplate.MatchInReplyTo(reservationCode);

				ACLMessage msgCFPReply = myAgent.receive(mtCFPReply);
				
				if (msgCFPReply != null) {

					//If our proposal is accepted, we make the reservation
					if (msgCFPReply.getPerformative()==ACLMessage.ACCEPT_PROPOSAL) {

						//The reservation will start and finish the same day
						Reservation r = new Reservation();
						r.setTimeStart(timeStart.getTime());
						r.setTimeEnd(timeEnd.getTime());
						r.setjourneyManagerId(journeyManagerId);
						if (startToday) todayReservations.list.add(r);
						else tomorrowReservations.list.add(r);
					}
					step++;
					break;
				} else block();
			}
			}

		@Override
		public boolean done() {
			if (step == 4) {
				return true;
			}
			return false;
		}

	}	


private class BParkingSpotManagerOccupationRequest extends Behaviour {

		int step = 0;
		ACLMessage msg;
		Calendar today = Calendar.getInstance();
		AID restArea;
		String occupationCode;
		Calendar timeStart, timeEnd;
		String journeyManagerId;
		Boolean occupiedSlot = false;
		Boolean startToday;
	
		public BParkingSpotManagerOccupationRequest(ACLMessage msg) {
			// TODO Auto-generated constructor stub
			this.msg = msg;
		}
	
		@Override
		public void action() {
			switch (step){
			case 0:
				today = Calendar.getInstance();
				//Reading the message
				restArea = msg.getSender();
				occupationCode = msg.getInReplyTo();
				try {
					ContentElement content = getContentManager().extractContent(msg);
					OccupationRequest occupationRequest = (OccupationRequest) ((Action) content).getAction();
					Reservation reservation = occupationRequest.getReservation();
					timeStart = Calendar.getInstance();
					timeEnd = Calendar.getInstance();
					timeStart.setTime(reservation.getTimeStart());
					timeEnd.setTime(reservation.getTimeEnd());
					journeyManagerId = reservation.getjourneyManagerId();
				} catch (CodecException | OntologyException e) {
					e.printStackTrace();
				}
				step++;
				break;
			case 1:
				//Check for free space
				startToday = (timeStart.get(Calendar.DAY_OF_YEAR) == today.get(Calendar.DAY_OF_YEAR));
				//The reservation will start today
				if (startToday) {
					//If the list is empty...
					if (todayReservations.list.isEmpty()) {
						step++;
						break;
					}
					//Check for all reservations for this day
					for ( Reservation reservation : todayReservations.list) {
						Calendar timeStartReservation = Calendar.getInstance();
						timeStartReservation.setTime(reservation.getTimeStart());
						Calendar timeEndReservation = Calendar.getInstance();
						timeEndReservation.setTime(reservation.getTimeEnd());
						//The asked time reservation will end in an occupied block of time
						if (timeStart.before(timeStartReservation) && timeEnd.after(timeStartReservation)){
							occupiedSlot = true;
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(occupationCode);
							myAgent.send(msgReplyRefuse);
							step++;
							break;
						}
						//The asked time reservation will start between an occupied block of time
						if ((timeStart.after(timeStartReservation) && timeStart.before(timeEndReservation)) || timeStart.equals(timeStartReservation)){
							occupiedSlot = true;
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(occupationCode);
							myAgent.send(msgReplyRefuse);
							step++;
							break;
						}
					}
				}

				//The reservation will begin tomorrow
				else {
					//Check for all reservations for today
					for ( Reservation reservation : todayReservations.list ) {
						Calendar timeStartReservation = Calendar.getInstance();
						timeStartReservation.setTime(reservation.getTimeStart());
						Calendar timeEndReservation = Calendar.getInstance();
						timeEndReservation.setTime(reservation.getTimeEnd());
						//The asked time reservation will end in an occupied block of time
						if (timeStart.before(timeStartReservation) && timeEnd.after(timeStartReservation)){
							occupiedSlot= true;
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(occupationCode);
							myAgent.send(msgReplyRefuse);
							step++;
							break;
						}
						//The asked time reservation will start between an occupied block of time
						if ((timeStart.after(timeStartReservation) && timeStart.before(timeEndReservation)) || timeStart.equals(timeStartReservation)){
							occupiedSlot= true;
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(occupationCode);
							myAgent.send(msgReplyRefuse);
							step++;
							break;
						}
					}
					//Check for all reservations for tomorrow
					for ( Reservation reservation : tomorrowReservations.list ) {
						Calendar timeStartReservation = Calendar.getInstance();
						timeStartReservation.setTime(reservation.getTimeStart());
						Calendar timeEndReservation = Calendar.getInstance();
						timeEndReservation.setTime(reservation.getTimeEnd());
						//The asked time reservation will end in an occupied block of time
						if (timeStart.before(timeStartReservation) && timeEnd.after(timeStartReservation)){
							occupiedSlot= true;
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(occupationCode);
							myAgent.send(msgReplyRefuse);
							step++;
							break;
						}
						//The asked time reservation will start between an occupied block of time
						if ((timeStart.after(timeStartReservation) && timeStart.before(timeEndReservation)) || (timeStart.equals(timeStartReservation))){
							occupiedSlot= true;
							ACLMessage msgReplyRefuse = new ACLMessage(ACLMessage.REFUSE);
							msgReplyRefuse.addReceiver(restArea);
							msgReplyRefuse.setInReplyTo(occupationCode);
							myAgent.send(msgReplyRefuse);
							step++;
							break;
						}
					}
				}
			case 2:
				if (!occupiedSlot) {
					ACLMessage msgReplyAgree = new ACLMessage(ACLMessage.AGREE);
					msgReplyAgree.addReceiver(restArea);
					msgReplyAgree.setInReplyTo(occupationCode);
					myAgent.send(msgReplyAgree);
					step++;
					break;
				}
			}
		}
	
		@Override
		public boolean done() {
			if (step == 3) return true;
			return false;
		}
	}


	private class BParkingSpotManagerCancelRequest extends Behaviour {

		int step = 0;
		ACLMessage msg;
		String journeyManagerId;
		AID restArea;
		
		public BParkingSpotManagerCancelRequest(ACLMessage msg) {
			// TODO Auto-generated constructor stub
			this.msg = msg;
		}
	
		@Override
		public void action() {
			switch (step){
			case 0:
				//Reading the message
				restArea = msg.getSender();
				try {
					ContentElement content = getContentManager().extractContent(msg);
					CancelRequest cancelRequest = (CancelRequest) ((Action) content).getAction();
					Reservation reservation = cancelRequest.getReservation();
					journeyManagerId = reservation.getjourneyManagerId();
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				step++;
				break;
			case 1:
				for ( Reservation reservation : todayReservations.list) {
					if (reservation.getjourneyManagerId() == journeyManagerId) {
						todayReservations.list.remove(reservation);
						step++;
						break;
						}
					}
				for ( Reservation reservation : tomorrowReservations.list ) {
					if (reservation.getjourneyManagerId() == journeyManagerId) {
						todayReservations.list.remove(reservation);
						step++;
						break;
						}
					}
				//The reservation doesn't exist
				step++;
				break;
			}
		}
	
		@Override
		public boolean done() {
			if (step == 2) return true;
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//	

	
	//******************************************************************************************//
	//******************************************************************************************//	
	// General behaviours: BParkingSpotShutDown													//
	//******************************************************************************************//
	//******************************************************************************************//
	private class BParkingSpotShutDown extends Behaviour {

		@Override
		public void action() {
			// TODO Auto-generated method stub
			MessageTemplate mtInformShutDown = MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.INFORM), MessageTemplate.MatchConversationId("shutDown"));

			ACLMessage msgInformUpdate = myAgent.receive(mtInformShutDown);
			if (msgInformUpdate != null) {
				myAgent.doDelete();
			}
		}

		@Override
		public boolean done() {
			// TODO Auto-generated method stub
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//	

}
