package agents;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Random;
import java.util.UUID;

import ontology.CancelRequest;
import ontology.LogJourneyData;
import ontology.LogRelocation;
import ontology.OccupationRequest;
import ontology.ParkingManagementOntology;
import ontology.Reservation;
import ontology.ReservationRequest;
import ontology.RestArea;
import ontology.Road;
import jade.content.Concept;
import jade.content.ContentElement;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.CyclicBehaviour;
import jade.domain.AMSService;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.wrapper.AgentController;
import jade.wrapper.ContainerController;
import jade.wrapper.StaleProxyException;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
//
import jade.content.onto.basic.*;


/**
 * @author Francisco Palau Romero
 * @code: Rest Area agent code 
 *
**/

public class AgRestAreaManager extends Agent {
	private Road road;
	private int km;
	private int numParkingSpots;
	private int numOccupiedParkingSpots;
	private int idRestArea;
	private AID ID;
	private ArrayList<AID> parkingSpotList = new ArrayList<AID>();
	//Simulation type
	private int simulationType;
	private static final int BORDA_NEGOTIATION_SIMULATION_TYPE = 0;
	private static final int FREE_SIMULATION_TYPE = 1;
	//Log AID
	private AID logManagerAID;
	//Ontology
	private Codec codec = new SLCodec();
	private Ontology ontoloy = ParkingManagementOntology.getInstance();
	Action action;

	//******************************************************************************************//
	//******************************************************************************************//	
	// Setup																					//
	//******************************************************************************************//
	//******************************************************************************************//	
	protected void setup() {
		//Register codec and ontology
	    getContentManager().registerLanguage(codec);
	    getContentManager().registerOntology(ontoloy);
	    //

	    //Extract parameters from Loader
		Object[] args = getArguments();
		RestArea restAreaInfo = (RestArea) args[0];
		simulationType = (Integer) args[1];
		road = restAreaInfo.getRoad();
		km = restAreaInfo.getKM();
		numParkingSpots = restAreaInfo.getAreaSpots();
		idRestArea = restAreaInfo.getRestAreaId();
		//

		//Register the offered service
		DFAgentDescription dfd = new DFAgentDescription();
		ID = getAID();
		dfd.setName(ID);
		ServiceDescription sd = new ServiceDescription();
		sd.setType("RestArea");
		sd.setName(road.getName());
		dfd.addServices(sd);
		try{
			DFService.register(this, dfd);
		} catch (Exception e) {e.printStackTrace();}
		//

		//Register the AID of the log manager
		DFAgentDescription[] result = null;
		ServiceDescription sdLog = new ServiceDescription(); 
		DFAgentDescription dfdLog = new DFAgentDescription();

		sdLog.setType("LogManager");
		sdLog.setName("logManager");
		dfdLog.addServices(sdLog);

		while (result == null) {
		    try {
		    	//Search for 100 ms
		    	result = DFService.searchUntilFound(this, getDefaultDF(), dfdLog, null, 100);
			} catch (FIPAException e) { 
				e.printStackTrace(); 
				System.out.println("FIPA Error while searching for agents");
			}
		}
		
	    if (result.length >= 1) {
	    	for (DFAgentDescription logManagerResult : result){
	    		logManagerAID = logManagerResult.getName();
	    	}
	    }
	    //


	    //The type of simulation it will depend of simulationType
	    switch (simulationType) {
		    case FREE_SIMULATION_TYPE:
				numOccupiedParkingSpots = 0;
				addBehaviour(new BRestAreaManagerSetSpotFree());
				addBehaviour(new BRestAreaManagerReservationFree());	    	
		    	break;
		    case BORDA_NEGOTIATION_SIMULATION_TYPE:
				//Create parking spots
				ContainerController cc = getContainerController();
				AgentController ac;
				for (int i=1;i<=numParkingSpots;i++) {
					try {
						String spotName = getAID().getLocalName()+"spot"+i;
						ac = cc.createNewAgent(spotName, "agents.AgParkingSpot", null);
						ac.start();
						parkingSpotList.add(new AID(spotName,false));
					} catch (StaleProxyException e) {
						e.printStackTrace();
					}
				}
	
				addBehaviour(new BRestAreaMessageReceiver());
				addBehaviour(new BRestAreaManagerReservationRequest());
		    	break;
	    }
	    //
	    
	    //Shut down behaviour
		addBehaviour(new BRestAreaManagerShutDown());
		//
	}
	//******************************************************************************************//
	//******************************************************************************************//	

	
	//******************************************************************************************//
	//******************************************************************************************//	
	// Override																					//
	//******************************************************************************************//
	//******************************************************************************************//
	@Override
	protected void takeDown() {
		while (receive() != null) {};
	}
	//******************************************************************************************//
	//******************************************************************************************//	

	
	//******************************************************************************************//
	//******************************************************************************************//
	// Borda Negotiation Behavious: BRestAreaMessageReceiver, BRestAreaManagerOccupationRequest,//
	//								BRestAreaManagerReservationRequest,							//
	//								BRestAreaManagerNegotiateForRelocation						//
	//******************************************************************************************//
	//******************************************************************************************//
	//This behaviour allows to create the behaviours needed 
	private class BRestAreaMessageReceiver extends CyclicBehaviour {

		public void action(){
			ACLMessage msg = receive(MessageTemplate.MatchConversationId("occupation"));
			if (msg==null){
				block();
				return;
			}

	       try {
	    	   ContentElement content = getContentManager().extractContent(msg);
		       Concept action = ((Action) content).getAction();
		       switch (msg.getPerformative()) { 
		       	case (ACLMessage.REQUEST):
		       		if (action instanceof OccupationRequest) {
		       			addBehaviour(new BRestAreaManagerOccupationRequest(msg));
		       		}
		       	break;
		       }
	       	}
	       	catch(Exception ex) { ex.printStackTrace(); }
		}
	}
	
	//This behaviour allos to create a reservation
	private class BRestAreaManagerReservationRequest extends Behaviour {

		int step = 0;
		MessageTemplate mtRequestReservation = MessageTemplate.MatchPerformative(ACLMessage.REQUEST);
		AID journeyManager;
		String replyCode;
		String msgContent;
		int numResp = 0;
		ArrayList<AID> freeParkingSpots = new ArrayList<AID>();
		ArrayList<AID> occupiedParkingSpots = new ArrayList<AID>();
		ArrayList<AID> driversToNegotiate = new ArrayList<AID>();
		int time = 0, bestTime = -999999;
		AID bestParkingSpot;
		Calendar timeStart, timeEnd;
		String journeyManagerId;
		

		@Override
		public void action() {
			switch (step){
			case 0:
				//Reading the message
				ACLMessage msg = receive(MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.REQUEST),MessageTemplate.MatchConversationId("request")));
				if (msg != null) {
					replyCode = msg.getInReplyTo();
					journeyManager = msg.getSender();
					try {
						ContentElement content = getContentManager().extractContent(msg);
						ReservationRequest reservationRequest = (ReservationRequest) ((Action) content).getAction();
						Reservation reservation = reservationRequest.getReservation();
						journeyManagerId = reservation.getjourneyManagerId();
						timeStart = Calendar.getInstance();
						timeEnd = Calendar.getInstance();
						timeStart.setTime(reservation.getTimeStart());
						timeEnd.setTime(reservation.getTimeEnd());
					} catch (CodecException | OntologyException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					step++;
				}
				else block();
				break;
			case 1:
				//Asking to parking spots if they're available
				ACLMessage msgCFPParkingSpot = new ACLMessage(ACLMessage.CFP);
				msgCFPParkingSpot.setLanguage(codec.getName());
				msgCFPParkingSpot.setOntology(ontoloy.getName());

				//Content of the msg
				Reservation reservation = new Reservation();
				reservation.setTimeStart(timeStart.getTime());
				reservation.setTimeEnd(timeEnd.getTime());
				reservation.setjourneyManagerId(journeyManagerId);
				ReservationRequest reservationRequest = new ReservationRequest();
				reservationRequest.setReservation(reservation);
				//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
				action = new Action(getAID(),reservationRequest);
				try {
					getContentManager().fillContent(msgCFPParkingSpot, action);
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

				for (AID parkingSpot : parkingSpotList) {
					msgCFPParkingSpot.addReceiver(parkingSpot);
				}
				msgCFPParkingSpot.setInReplyTo(replyCode);
				msgCFPParkingSpot.setConversationId("request");
				myAgent.send(msgCFPParkingSpot);
				step++;
				break;
			case 2:
				//Checking if some parking spot is available
				MessageTemplate mtCFPReply = MessageTemplate.MatchInReplyTo(replyCode);

				ACLMessage msgCFPReply = myAgent.receive(mtCFPReply);
				
				if (msgCFPReply != null) {
					if (msgCFPReply.getPerformative()==ACLMessage.PROPOSE) {
						freeParkingSpots.add(msgCFPReply.getSender());
						time = Integer.parseInt(msgCFPReply.getContent());

						if (time > bestTime) {
							bestTime = time;
							bestParkingSpot = msgCFPReply.getSender();
						}
					}
					//The ones who doesn't have free spot, would reply with the local name of the journey manager agent involved
					else {
						driversToNegotiate.add(new AID(msgCFPReply.getContent(),AID.ISLOCALNAME));
						occupiedParkingSpots.add(msgCFPReply.getSender());
					}
					numResp++;
					if (numResp>=parkingSpotList.size()) step++;
				} else block();
				break;
			case 3:
				//If there is a parking spot available, accept the proposal
				if (bestParkingSpot!= null) {
					ACLMessage response = new ACLMessage(ACLMessage.REJECT_PROPOSAL);
					response.setInReplyTo(replyCode);
					
					//Send reject messages
					for (AID parkingSpot : parkingSpotList) 
						if (!parkingSpot.getLocalName().equals(bestParkingSpot.getLocalName())) 
							response.addReceiver(parkingSpot);
					myAgent.send(response);
					
					//Send accept message
					response = new ACLMessage(ACLMessage.ACCEPT_PROPOSAL);
					response.setInReplyTo(replyCode);
					response.addReceiver(bestParkingSpot);
					response.setContent(journeyManager.toString());
					myAgent.send(response);
				   
					//Tell to the Journey Manager that his reservation has been done
					response = new ACLMessage(ACLMessage.AGREE);
					response.setInReplyTo(replyCode);
					response.addReceiver(journeyManager);
					myAgent.send(response);
				}
				else {
					//If there is not an available parking spot, the negotiation behavior will start
					addBehaviour(new BRestAreaManagerNegotiateForRelocation(journeyManager, replyCode, msgContent, driversToNegotiate, occupiedParkingSpots));
				}
			step++;
			break;
			}
		}

		@Override
		public boolean done() {
			if (step == 4) {
				addBehaviour(new BRestAreaManagerReservationRequest());
				return true;
			}
			return false;
		}

	}

	//This behaviour return if has or hasen't any free parking spot
	private class BRestAreaManagerOccupationRequest extends Behaviour {

		int step = 0;
		MessageTemplate mtRequestOccupation = MessageTemplate.MatchPerformative(ACLMessage.REQUEST);
		AID journeyManager;
		String replyCode;
		String journeyManagerId;
		Calendar timeStart, timeEnd;
		int numResp = 0;
		ACLMessage msg;
		boolean freeSpot;
		
		
		public BRestAreaManagerOccupationRequest(ACLMessage msg) {
			// TODO Auto-generated constructor stub
			this.msg = msg;
		}

		@Override
		public void action() {
			switch (step){
			case 0:
				replyCode = msg.getInReplyTo();
				journeyManager = msg.getSender();
				try {
					ContentElement content = getContentManager().extractContent(msg);
					OccupationRequest occupationRequest = (OccupationRequest) ((Action) content).getAction();
					Reservation reservation = occupationRequest.getReservation();
					journeyManagerId = reservation.getjourneyManagerId();
					timeStart = Calendar.getInstance();
					timeEnd = Calendar.getInstance();
					timeStart.setTime(reservation.getTimeStart());
					timeEnd.setTime(reservation.getTimeEnd());
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				step++;
				break;
			case 1:
				//Asking to parking spots if they're available
				ACLMessage msgCFPParkingSpot = new ACLMessage(ACLMessage.REQUEST);
				for (AID parkingSpot : parkingSpotList) {
					msgCFPParkingSpot.addReceiver(parkingSpot);
				}

				msgCFPParkingSpot.setLanguage(codec.getName());
				msgCFPParkingSpot.setOntology(ontoloy.getName());
				
				//Content of the msg
				Reservation reservation = new Reservation();
				reservation.setTimeStart(timeStart.getTime());
				reservation.setTimeEnd(timeEnd.getTime());
				reservation.setjourneyManagerId(journeyManagerId);
				OccupationRequest occupationRequest = new OccupationRequest();
				occupationRequest.setReservation(reservation);
				//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
				action = new Action(getAID(),occupationRequest);
				try {
					getContentManager().fillContent(msgCFPParkingSpot, action);
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

				msgCFPParkingSpot.setInReplyTo("occupation");
				msgCFPParkingSpot.setConversationId("request");
				myAgent.send(msgCFPParkingSpot);
				step++;
				break;
			case 2:
				//Checking if some parking spot is available
				MessageTemplate mtCFPReply = MessageTemplate.MatchInReplyTo("occupation");
				ACLMessage msg_cfp_reply = myAgent.receive(mtCFPReply);
				freeSpot = false;
				
				if (msg_cfp_reply != null) {
					if (msg_cfp_reply.getPerformative()==ACLMessage.AGREE) {
						freeSpot = true;
					}
					numResp++;
					if (numResp>=parkingSpotList.size()) step++;
				} else block();
				break;
			case 3:
				//If there is a parking spot available, accept the proposal
				ACLMessage response;
				if (freeSpot) response = new ACLMessage(ACLMessage.AGREE);
				else response = new ACLMessage(ACLMessage.REFUSE);
				//Send accept message
				response.setInReplyTo(replyCode);
				response.addReceiver(journeyManager);
				myAgent.send(response);
				step++;
				break;
			}
		}

		@Override
		public boolean done() {
			if (step == 4) return true;
			return false;
		}

	}


	private class BRestAreaManagerNegotiateForRelocation extends Behaviour {

		int step = 0;
		int numResp = 0;
		MessageTemplate mtCFPReply;
		MessageTemplate mtConfirmReservation;
		ArrayList<AID> driversToNegotiate = new ArrayList<AID>();
		ArrayList<AID> occupiedParkingSpots = new ArrayList<AID>();
		ArrayList<AID> driversToChange = new ArrayList<AID>();
		ArrayList<Float> proposedTimes = new ArrayList<Float>();
		AID bestDriver;
		float bestChange = 99999;
		AID journeyManager;
		String replyCode;
		String msgContent;
		boolean newReservationCompleted = false;
		boolean waitingForResponse = false;
		//Data for the log
		int affectedVehicles = 0;
		int relocableVehicles = 0;
		float relocatedVehiclesLostTime = 0;
		int relocatedVehicleOrder = 0;
		Calendar relocationTime = Calendar.getInstance();
		boolean success = false;


		public BRestAreaManagerNegotiateForRelocation(AID journeyManager1, String replyCode1, String msgContent1, ArrayList<AID> driversToNegotiate1, ArrayList<AID> occupiedParkingSpots1) {
			journeyManager = journeyManager1;
			replyCode = replyCode1;
			msgContent = msgContent1;
			driversToNegotiate = driversToNegotiate1;
			affectedVehicles = driversToNegotiate.size();
			occupiedParkingSpots = occupiedParkingSpots1;
		}

		@Override
		public void action() {
			switch (step){
			case 0:
				ACLMessage msgCFPDriverToNegotiate = new ACLMessage(ACLMessage.CFP);
				//Including journeyManager to the negotiation
				driversToNegotiate.add(journeyManager);
				occupiedParkingSpots.add(null); //The journey manager who ask for a place, hasn't parking spot yet
				for (AID driver : driversToNegotiate) {
					msgCFPDriverToNegotiate.addReceiver(driver);
				}
				msgCFPDriverToNegotiate.setContent(getAID().toString());
				msgCFPDriverToNegotiate.setInReplyTo(replyCode);
				myAgent.send(msgCFPDriverToNegotiate);
				
				step++;
				break;
			case 1:
				//Checking the drivers avaliable
				mtCFPReply = MessageTemplate.MatchInReplyTo(replyCode);

				ACLMessage msgCFPDriverReply = myAgent.receive(mtCFPReply);

				if (msgCFPDriverReply != null) {
					if (msgCFPDriverReply.getPerformative()==ACLMessage.PROPOSE) {
						relocableVehicles++;
						driversToChange.add(msgCFPDriverReply.getSender());
						float time = Float.parseFloat(msgCFPDriverReply.getContent());
						proposedTimes.add(time);
						if (time < bestChange) {
							bestChange = time;
							bestDriver = msgCFPDriverReply.getSender();
						}
					}
					numResp++;
					if (numResp>=driversToNegotiate.size()) step++;
				} else block();
				break;
			case 2:
				//We got at least 1 driver willing to change their occupation
				if (bestDriver!= null) {
					//If the message for the new reservation hasn't been sent...
					if (!waitingForResponse && !newReservationCompleted) {
						ACLMessage response = new ACLMessage(ACLMessage.ACCEPT_PROPOSAL); 
						//Send accept message
						response.setInReplyTo(replyCode);
						response.addReceiver(bestDriver);
						myAgent.send(response);
						relocatedVehicleOrder++;
						waitingForResponse = true;
					}
					//If the message for the new reservation has been sent
					if (waitingForResponse && !newReservationCompleted)  {
						mtConfirmReservation = MessageTemplate.MatchInReplyTo(replyCode);

						ACLMessage msgCFPDriverReservation = myAgent.receive(mtConfirmReservation);

						if (msgCFPDriverReservation != null) {
							//If everything is ok, the parking spot must be freed
							if (msgCFPDriverReservation.getPerformative() == ACLMessage.INFORM){
								newReservationCompleted = true;
								int index = driversToNegotiate.indexOf(bestDriver);
								ACLMessage msgCancellation = new ACLMessage(ACLMessage.REQUEST);
								msgCancellation.addReceiver(occupiedParkingSpots.get(index));
								msgCancellation.setLanguage(codec.getName());
								msgCancellation.setOntology(ontoloy.getName());

								//Content of the msg
								Reservation reservation = new Reservation();
								reservation.setjourneyManagerId(bestDriver.getLocalName());

								Calendar foo = Calendar.getInstance();
								reservation.setTimeStart(foo.getTime());
								reservation.setTimeEnd(foo.getTime());
								CancelRequest cancelRequest = new CancelRequest();
								cancelRequest.setReservation(reservation);
								//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
								action = new Action(getAID(),cancelRequest);
								try {
									getContentManager().fillContent(msgCancellation, action);
								} catch (CodecException | OntologyException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								}

								myAgent.send(msgCancellation);
							}
							//If not, we repeat the process
							else {
								//If this was the last driver, the negotiation fails
								if (driversToChange.isEmpty()) {
									break;
								}
								//Remove the driver that couldn't make the reservation
								int index = driversToChange.indexOf(bestDriver);
								proposedTimes.remove(index);
								driversToChange.remove(index);
								
								//Checking for the new best time
								float time = 99999;
								int bestIndex = 0;
								index = 0;

								for (float newBestTime : proposedTimes) {
									if (time > newBestTime) {
										time = newBestTime;
										bestIndex = index;
									}
									index++;
								}
								bestChange = time;
								
								bestDriver = driversToChange.get(bestIndex);
								waitingForResponse = false;
							}
						} else block();
					}

					//Send reject message
					if (newReservationCompleted) {
						ACLMessage rejectDrivers = new ACLMessage(ACLMessage.REJECT_PROPOSAL);
						rejectDrivers.setInReplyTo(replyCode);
						for (AID parkingSpot : driversToChange) 
							if (!parkingSpot.getLocalName().equals(bestDriver.getLocalName())) 
								rejectDrivers.addReceiver(parkingSpot);
						myAgent.send(rejectDrivers);
						relocatedVehiclesLostTime = bestChange;
						success = true;
						step++;
					}
				}
				else {
					//If there is not an available parking spot, the negotiation is over and the driver that initiated the negotiation must seek for another rest area
					ACLMessage response = new ACLMessage(ACLMessage.REFUSE);
					response.setInReplyTo(replyCode);
					response.addReceiver(journeyManager);
					myAgent.send(response);
					success = false;
					step++;
					break;
				}
			}
		}

		@Override
		public boolean done() {
			if (step == 3){
				ACLMessage msgLogRelocation = new ACLMessage(ACLMessage.INFORM);
				msgLogRelocation.setLanguage(codec.getName());
				msgLogRelocation.setOntology(ontoloy.getName());

				LogRelocation logRelocation = new LogRelocation();

				logRelocation.setAffectedVehicles(affectedVehicles);
				logRelocation.setRelocableVehicles(relocableVehicles);
				logRelocation.setRelocatedVehicleLostTime(relocatedVehiclesLostTime);
				logRelocation.setRelocatedVehicleOrder(relocatedVehicleOrder);
				logRelocation.setRelocationTime(relocationTime.getTime());
				logRelocation.setSuccess(success);

				//The action must be "stored" in an standard "Action" object, instead of a "custom" AgentAction
				action = new Action(getAID(),logRelocation);
				try {
					getContentManager().fillContent(msgLogRelocation, action);
				} catch (CodecException | OntologyException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				msgLogRelocation.setConversationId("log");
				msgLogRelocation.addReceiver(logManagerAID);
				send(msgLogRelocation);

				return true;
			}
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//	

	
	//******************************************************************************************//
	//******************************************************************************************//	
	// Simulation without negotiation behaviour: BRestAreaManagerReservationFree				//
	//											 BRestAreaManagerSetSpotFree					//
	//******************************************************************************************//
	//******************************************************************************************//	
	private class BRestAreaManagerReservationFree extends Behaviour {

		int step = 0;
		AID journeyManager;
		String replyCode;

		@Override
		public void action() {
			switch (step){
			case 0:
				//Reading the message
				ACLMessage msg = receive(MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.REQUEST),MessageTemplate.MatchConversationId("requestFree")));
				if (msg != null) {
					replyCode = msg.getInReplyTo();
					journeyManager = msg.getSender();
					step++;
				}
				else block();
				break;
			case 1:
				//Checking if some parking spot is available
				if (numOccupiedParkingSpots < numParkingSpots) {
					ACLMessage msgResponse = new ACLMessage(ACLMessage.AGREE);
					msgResponse.setInReplyTo(replyCode);
					msgResponse.addReceiver(journeyManager);
					numOccupiedParkingSpots++;
					myAgent.send(msgResponse);
				}
				else {
					ACLMessage msgResponse = new ACLMessage(ACLMessage.REFUSE);
					msgResponse.setInReplyTo(replyCode);
					msgResponse.addReceiver(journeyManager);
					myAgent.send(msgResponse);
				}
				step++;
				break;
			}
		}

		@Override
		public boolean done() {
			if (step == 2) {
				addBehaviour(new BRestAreaManagerReservationFree());
				return true;
			}
			return false;
		}
	}
		
	private class BRestAreaManagerSetSpotFree extends Behaviour {
		int step = 0;

		@Override
		public void action() {
			switch (step){
			case 0:
				//Reading the message
				ACLMessage msg = receive(MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.REQUEST),MessageTemplate.MatchConversationId("requestSetSpotFree")));
				if (msg != null) {
					numOccupiedParkingSpots--;
					step++;
					break;
				}
			}
		}

		@Override
		public boolean done() {
			if (step == 1) {
				addBehaviour(new BRestAreaManagerSetSpotFree());
				return true;
			}
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//	


	//******************************************************************************************//
	//******************************************************************************************//	
	// General behaviours: BRestAreaManagerShutDown												//
	//******************************************************************************************//
	//******************************************************************************************//	
	private class BRestAreaManagerShutDown extends Behaviour {

		@Override
		public void action() {
			// TODO Auto-generated method stub
			MessageTemplate mtInformShutDown = MessageTemplate.and(MessageTemplate.MatchPerformative(ACLMessage.INFORM), MessageTemplate.MatchConversationId("shutDown"));

			ACLMessage msgInformUpdate = myAgent.receive(mtInformShutDown);
			if (msgInformUpdate != null) {
					myAgent.doDelete();
				}
			}

		@Override
		public boolean done() {
			// TODO Auto-generated method stub
			return false;
		}
	}
	//******************************************************************************************//
	//******************************************************************************************//	

}
