package gui;

import jade.gui.GuiEvent;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JRadioButton;
import javax.swing.JSeparator;
import javax.swing.JTextField;
import javax.swing.SwingConstants;

import agents.AgLoader;

import javax.swing.JComboBox;
import javax.swing.DefaultComboBoxModel;

import com.ibm.icu.impl.duration.TimeUnit;

import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.Date;

import javax.swing.JProgressBar;

/**
 * @author Raquel Garca Arvalo
 * @code: Loader Graphic User Interface
 * 
 */

public class LoaderGUI extends JFrame implements ActionListener, KeyListener {

	protected GUIData result;
	private agents.AgLoader myAgent;

	private JTextField setupFilesFolder;

	private JLabel lblSetupFolder;
	private JButton btnSelectFile;
	
	private JSeparator sepSimulation;
	private JLabel lblRealTimeMinutes;
	private JLabel lblSimulationTime;
	private JLabel lblRealTimeTo;
	private JLabel lblUpdateSimulationEach;
	private JLabel lblVehicles;
	private JLabel lblSeconds;
	private JLabel lblHeuristicType;
	private JLabel lblSeconds_1;
	private JLabel lbltxt;
	private JSeparator sepVehicles;
	private JLabel lblVehicleCreation;
	
	private JComboBox simulationType;
	private JTextField simulationTime;
	private JComboBox simTimeUnit;
	private JTextField simulationUpdate;
	private JTextField minutesPerLapse;
	private JLabel lbEstimation;

	private JTextField numberOfVehicles;
	private JTextField vehicleCreationLapse;

	private JRadioButton rdbtnFromRandomData, rdbtnFromFile;
	private JCheckBox chckbxCreateLog;
	private JTextField txLogName;

	private JButton btnStart, btnExit;
	
	private JProgressBar progressBar;
	private JLabel lblStatus;

	public void startSimulation() {
		simulationType.setVisible(false);
		setupFilesFolder.setVisible(false);
		simulationUpdate.setVisible(false);
		minutesPerLapse.setVisible(false);
		simulationTime.setVisible(false);	
		simTimeUnit.setVisible(false);	
		numberOfVehicles.setVisible(false);
		vehicleCreationLapse.setVisible(false);
		rdbtnFromFile.setVisible(false);
		rdbtnFromRandomData.setVisible(false);
		chckbxCreateLog.setVisible(false);
		txLogName.setVisible(false);
		lblSeconds_1.setVisible(false);
		lbltxt.setVisible(false);
		sepVehicles.setVisible(false);
		lblVehicleCreation.setVisible(false);
		
		lblSetupFolder.setVisible(false);
		setupFilesFolder.setVisible(false);
		
		btnSelectFile.setVisible(false);
		
		sepSimulation.setVisible(false);		
		lblRealTimeMinutes.setVisible(false);
		lblSimulationTime.setVisible(false);
		lblRealTimeTo.setVisible(false);
		lblUpdateSimulationEach.setVisible(false);
		lblVehicles.setVisible(false);
		lblSeconds.setVisible(false);
		lblHeuristicType.setVisible(false);
		
		btnStart.setVisible(false);
		
		getContentPane().add(progressBar);
		getContentPane().add(lblStatus);
		
	}

	public LoaderGUI(AgLoader loader) {
		setBackground(Color.LIGHT_GRAY);
		setForeground(Color.BLACK);

		myAgent = loader;
		result = new GUIData();

		setTitle("Parking Management Simulation Setup");
		getContentPane().setLayout(null);
		setResizable(false);
		setSize(new Dimension(500, 450));

		// SELECT SETUP FILES FOLDER
		lblSetupFolder = new JLabel("Setup files folder:");
		lblSetupFolder.setToolTipText("Agent data will be loaded from files at this location. The name and fields for the files are explained at documentation.");
		lblSetupFolder.setBounds(20, 15, 124, 14);
		getContentPane().add(lblSetupFolder);
		
		setupFilesFolder = new JTextField();
		setupFilesFolder.setToolTipText("Agent data will be loaded from files at this location. The name and fields for the files are explained at documentation.");
		setupFilesFolder.setText("SetupFiles");
		setupFilesFolder.setBounds(147, 11, 302, 20);
		getContentPane().add(setupFilesFolder);
		setupFilesFolder.setColumns(10);

		btnSelectFile = new JButton("...");
		btnSelectFile.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent arg0) {

				JFileChooser j = new JFileChooser();
				j.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				Integer option = j.showOpenDialog(null);

				if (option == JFileChooser.APPROVE_OPTION) {
					setupFilesFolder.setText(j.getSelectedFile().toString());

				}

			}
		});
		btnSelectFile.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
			}
		});
		btnSelectFile.setBounds(460, 11, 24, 23);
		getContentPane().add(btnSelectFile);

		// SIMULATION SETUP

		lblSimulationTime = new JLabel("Simulation Parameters");
		lblSimulationTime.setBounds(20, 54, 144, 14);
		getContentPane().add(lblSimulationTime);
		simulationTime = new JTextField();

		
		simulationTime.setHorizontalAlignment(SwingConstants.RIGHT);
		simulationTime.setText("1");
		simulationTime.setBounds(192, 113, 69, 20);
		getContentPane().add(simulationTime);
		simulationTime.setColumns(10);

		simulationTime.addKeyListener(this);		

		sepSimulation = new JSeparator();
		sepSimulation.setBounds(134, 247, 360, 2);
		getContentPane().add(sepSimulation);

		lblRealTimeMinutes = new JLabel("Simulated minutes per lapse:");
		lblRealTimeMinutes.setBounds(38, 172, 180, 14);
		getContentPane().add(lblRealTimeMinutes);
		
		lblRealTimeTo = new JLabel("Time to simulate:");
		lblRealTimeTo.setBounds(38, 113, 152, 20);
		getContentPane().add(lblRealTimeTo);

		simulationUpdate = new JTextField();
		simulationUpdate.setHorizontalAlignment(SwingConstants.RIGHT);
		simulationUpdate.setText("1000");
		simulationUpdate.setBounds(192, 141, 69, 20);
		getContentPane().add(simulationUpdate);
		simulationUpdate.setColumns(10);
		
		simulationUpdate.addKeyListener(this);
		
		lblUpdateSimulationEach = new JLabel("Update simulation lapse:");
		lblUpdateSimulationEach.setBounds(38, 144, 152, 14);
		getContentPane().add(lblUpdateSimulationEach);
		
			

		minutesPerLapse = new JTextField();
		minutesPerLapse.setHorizontalAlignment(SwingConstants.RIGHT);
		minutesPerLapse.setText("15");
		minutesPerLapse.setBounds(218, 169, 43, 20);
		getContentPane().add(minutesPerLapse);
		minutesPerLapse.setColumns(10);

		minutesPerLapse.addKeyListener(this);
		
		lblSeconds_1 = new JLabel("ms");
		lblSeconds_1.setBounds(271, 141, 43, 20);
		getContentPane().add(lblSeconds_1);

		lbltxt = new JLabel(".txt");
		lbltxt.setBounds(328, 321, 29, 14);
		getContentPane().add(lbltxt);

		// VEHICLES CREATION SETUP

		sepVehicles = new JSeparator();
		sepVehicles.setBounds(160, 66, 320, 2);
		getContentPane().add(sepVehicles);

		lblVehicleCreation = new JLabel("Vehicle Creation");
		lblVehicleCreation.setBounds(20, 235, 104, 14);
		getContentPane().add(lblVehicleCreation);

		rdbtnFromFile = new JRadioButton("From Vehicles setup file");
		rdbtnFromFile.setSelected(true);
		rdbtnFromFile.setBounds(36, 256, 386, 23);
		getContentPane().add(rdbtnFromFile);
		rdbtnFromFile.addActionListener(this);

		rdbtnFromRandomData = new JRadioButton("Create random");
		rdbtnFromRandomData.setBounds(36, 282, 128, 23);
		getContentPane().add(rdbtnFromRandomData);
		rdbtnFromRandomData.addActionListener(this);

		ButtonGroup bG = new ButtonGroup();
		bG.add(rdbtnFromFile);
		bG.add(rdbtnFromRandomData);

		numberOfVehicles = new JTextField();
		numberOfVehicles.setHorizontalAlignment(SwingConstants.RIGHT);
		numberOfVehicles.setEnabled(false);
		numberOfVehicles.setBounds(173, 283, 43, 20);
		getContentPane().add(numberOfVehicles);
		numberOfVehicles.setColumns(10);

		lblVehicles = new JLabel("vehicles every");
		lblVehicles.setBounds(226, 286, 94, 14);
		getContentPane().add(lblVehicles);

		vehicleCreationLapse = new JTextField();
		vehicleCreationLapse.setHorizontalAlignment(SwingConstants.RIGHT);
		vehicleCreationLapse.setEnabled(false);
		vehicleCreationLapse.setBounds(323, 283, 43, 20);
		getContentPane().add(vehicleCreationLapse);
		vehicleCreationLapse.setColumns(10);

		lblSeconds = new JLabel("seconds");
		lblSeconds.setBounds(376, 286, 100, 14);
		getContentPane().add(lblSeconds);

		chckbxCreateLog = new JCheckBox("Create Vehicle File:");
		chckbxCreateLog.setBounds(55, 317, 144, 23);
		chckbxCreateLog.setEnabled(false);
		getContentPane().add(chckbxCreateLog);
		chckbxCreateLog.addActionListener(this);

		txLogName = new JTextField();
		txLogName.setText("VehicleFile");
		txLogName.setHorizontalAlignment(SwingConstants.RIGHT);
		txLogName.setEnabled(false);
		txLogName.setBounds(216, 318, 104, 20);
		txLogName.setEnabled(false);
		getContentPane().add(txLogName);
		txLogName.setColumns(10);

		lblHeuristicType = new JLabel("Heuristic type:");
		lblHeuristicType.setBounds(38, 84, 106, 14);
		getContentPane().add(lblHeuristicType);
		
		simulationType = new JComboBox();
		simulationType.setModel(new DefaultComboBoxModel(new String[] {"Negotiation", "Free Mode"}));
		simulationType.setSelectedIndex(0);
		simulationType.setEditable(true);
		simulationType.setBounds(147, 80, 114, 22);
		getContentPane().add(simulationType);
		
	
		simTimeUnit = new JComboBox();
		simTimeUnit.setModel(new DefaultComboBoxModel(new String[] {"minute(s)", "hour(s)", "day(s)", "year(s)"}));
		simTimeUnit.setSelectedIndex(3);
		simTimeUnit.setBounds(271, 113, 95, 21);
		getContentPane().add(simTimeUnit);
		simTimeUnit.addActionListener(this);

		
		//Calculate how much time it will take to simulate
		
		lbEstimation = new JLabel("");
		lbEstimation.setBounds(38, 208, 446, 14);
		getContentPane().add(lbEstimation);		
		updateEstimation();
		
	
		
		// BUTTONS
		// Start Simulation
		btnStart = new JButton("Start Simulation");
		btnStart.setBounds(140, 384, 160, 25);
		getContentPane().add(btnStart);
		btnStart.addActionListener(this);

		// Exit Simulation
		btnExit = new JButton("Exit");
		btnExit.setBounds(320, 384, 160, 25);
		getContentPane().add(btnExit);			

		btnExit.addActionListener(this);		
		

	}
	
	public void setStatusProgress(int value, Date date)
	{
		progressBar.setValue(value);
		lblStatus.setText("Simulation in progress. Simulation Time: "+date.toString());
	}

	private int simulationMinutes()
	{
		int minutes = 0;
		int totalTime = 0;
		try {
			totalTime = Integer.parseInt(simulationTime.getText());
		}
		catch (NumberFormatException nfe) {
			totalTime = 0;
		}
		
		switch(simTimeUnit.getSelectedIndex())
		{
		//minutes
		case (0):
			minutes = totalTime;
			break;
			
		//hours
		case(1):
			minutes = totalTime * 60;
			break;
			
		//days
		case(2):
			minutes = totalTime * 24 * 60;
			break;
			
		//years
		case(3):
			minutes = totalTime * 24 * 60 * 365;
			break;
			
		}
		
		return minutes;
	}
	
	
	private void updateEstimation() {
		int estimation = 0;		
		int update = Integer.parseInt(simulationUpdate.getText());
		int lapseTime = Integer.parseInt(minutesPerLapse.getText());
		
		try {
		//Calculate estimated simulation time
		estimation =  (update * simulationMinutes()) / lapseTime;
		
		//estimation now is ms, convert to minutes		
		estimation = (estimation / 1000) / 60; 
		
		if (estimation < 0)
			lbEstimation.setText("It will take less than a minute to simulate.");
		if (estimation < 60)			
			lbEstimation.setText("It will take around " + String.valueOf(estimation) + "  real time minutes to simulate.");
		else if (estimation >= 60)
			lbEstimation.setText("It will take around " + String.valueOf(estimation/60) + "  real time hours to simulate.");
		
		}
		catch (Error e)
		{
			lbEstimation.setText("It wasn't possible to estimate how long the simulation will take.");
		}
			
	}

	public void setSetupdata() {
		result.i_setupFilesFolder = setupFilesFolder.getText();
		
		result.i_simulationType = simulationType.getSelectedIndex();
		
		result.i_simulationTime = simulationMinutes();

		try {
			result.i_simMinutesPerLapse = Integer.parseInt(minutesPerLapse
					.getText());

		} catch (NumberFormatException nfe) {
			result.i_simMinutesPerLapse = 1;
		}

		try {
			result.i_simulationUpdate = Integer.parseInt(simulationUpdate
					.getText());

		} catch (NumberFormatException nfe) {
			result.i_simulationUpdate = 1000;
		}
		try {
			result.i_numberOfVehicles = Integer.parseInt(numberOfVehicles
					.getText());

		} catch (NumberFormatException nfe) {
			result.i_numberOfVehicles = 0;
		}

		try {
			result.i_vehicleCreationLapse = Integer
					.parseInt(vehicleCreationLapse.getText());

		} catch (NumberFormatException nfe) {
			result.i_vehicleCreationLapse = 1;
		}

		result.i_vehicleLogName = txLogName.getText();

		result.i_CreateLog = chckbxCreateLog.isSelected();

		if (rdbtnFromRandomData.isSelected())
			result.i_createRandomVehicles = true;
	}

	@Override
	public void actionPerformed(ActionEvent ae) {
		if (ae.getSource() == btnExit) {
			GuiEvent ge = new GuiEvent(this, 0);
			myAgent.postGuiEvent(ge);

			System.out.println("Closing...");
			System.exit(0);
		} else if (ae.getSource() == rdbtnFromRandomData
				|| ae.getSource() == rdbtnFromFile) {

			vehicleCreationLapse.setEnabled(rdbtnFromRandomData.isSelected());
			numberOfVehicles.setEnabled(rdbtnFromRandomData.isSelected());
			chckbxCreateLog.setEnabled(rdbtnFromRandomData.isSelected());
			txLogName.setEnabled(chckbxCreateLog.isSelected()
					&& chckbxCreateLog.isEnabled());

		} else if (ae.getSource() == chckbxCreateLog) {
			txLogName.setEnabled(chckbxCreateLog.isSelected());
		} else if (ae.getSource() == btnStart) {
			progressBar = new JProgressBar();
			progressBar.setBounds(20, 54, 464, 31);
			progressBar.setMaximum(simulationMinutes());
					
			
			lblStatus = new JLabel("Simulation in progress.");
			lblStatus.setBounds(20, 15, 464, 14);
				
			startSimulation();
			setSetupdata();
			btnExit.setText("Stop Simulation");

			GuiEvent ge = new GuiEvent(this, 1);
			ge.addParameter(result);
			myAgent.postGuiEvent(ge);
		} else if (ae.getSource() == simTimeUnit)
		{
			updateEstimation();
		}
		
		

	}
	
	@Override	
	 public void keyTyped(KeyEvent evt) 
	 {
		 if (evt.getSource() == simulationTime || evt.getSource() == simulationUpdate || evt.getSource() == minutesPerLapse)
		 {
			if (simulationTime.getText() == null && simulationTime.getText() == "")	
				 lbEstimation.setText("");
			else
				
				updateEstimation();
			
				

		 }
	 }

	@Override
	public void keyPressed(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void keyReleased(KeyEvent arg0) {
			
		updateEstimation();	
	}
}
