package ontology;

import jade.content.onto.BasicOntology;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.schema.AgentActionSchema;
import jade.content.schema.ConceptSchema;
import jade.content.schema.PrimitiveSchema;

import java.sql.Driver;

/**
 * @author Raquel Garca Arvalo
 * @code: ontology definition
 * 
 */

public class ParkingManagementOntology extends Ontology {

	public static final String ONTOLOGY_NAME = "Parking Management Ontology";

	// Vocabulary
	// Road
	public static final String ROAD = "Road";
	public static final String ROAD_ID = "roadId";
	public static final String ROAD_NAME = "name";
	public static final String ROAD_TRAFICC = "traffic";
	public static final String ROAD_LENGHT = "lenght";

	// Driver
	public static final String DRIVER = "Driver";
	public static final String DRIVER_ID = "driverId";
	public static final String DRIVER_NAME = "name";
	public static final String DRIVER_DRIVING_TIME = "drivingTime";

	// Vehicle
	public static final String VEHICLE = "Vehicle";
	public static final String VEHICLE_ID = "vehicleId";
	public static final String VEHICLE_MAXSPEED = "maxSpeed";
	public static final String VEHICLE_ROAD = "road";
	public static final String VEHICLE_KM = "KM";
	public static final String VEHICLE_DRIVER = "driver";
	// Potential improvement: Consider multiple drivers

	// RestArea
	public static final String REST_AREA = "Rest Area";
	public static final String REST_AREA_ID = "restAreaId";
	public static final String REST_AREA_NAME = "name";
	public static final String REST_AREA_SPOTS = "areaSpots";
	public static final String REST_AREA_ROAD = "road";
	public static final String REST_AREA_KM = "KM";

	// Reservation
	public static final String RESERVATION = "Reservation";
	public static final String RESERVATION_TIME_START = "timeStart";
	public static final String RESERVATION_TIME_END = "timeEnd";
	public static final String RESERVATION_JOURNEY_MANAGER_ID = "journeyManagerId";

	// Reservation Request
	public static final String RESERVATION_REQUEST = "ReservationRequest";
	public static final String RESERVATION_REQUEST_RESERVATION = "reservation";

	// Occupation request
	public static final String OCCUPATION_REQUEST = "OccupationRequest";
	public static final String OCCUPATION_REQUEST_RESERVATION = "reservation";

	// Cancel request
	public static final String CANCEL_REQUEST = "CancelRequest";
	public static final String CANCEL_REQUEST_RESERVATION = "reservation";

	//Log Journey Data
	public static final String LOG_JOURNEY_DATA = "LogJourneyData";
	public static final String LOG_JOURNEY_ID_VEHICLE = "idVehicle";
	public static final String LOG_JOURNEY_DATA_DRIVED_TIME = "drivedTime";
	public static final String LOG_JOURNEY_DATA_LOST_TIME = "lostTime";
	public static final String LOG_JOURNEY_DATA_REST_TIME = "restTime";
	public static final String LOG_JOURNEY_DATA_KM = "km";
	public static final String LOG_JOURNEY_DATA_KM_IN = "kmIn";
	public static final String LOG_JOURNEY_DATA_KM_OUT = "kmOut";
	public static final String LOG_JOURNEY_DATA_TIME_IN = "timeIn";
	public static final String LOG_JOURNEY_DATA_TIME_OUT = "timeOut";
	public static final String LOG_JOURNEY_INFRACTION_TYPE = "infractionType";

	//Log Reubication 
	public static final String LOG_RELOCATION = "LogRelocation";
	public static final String LOG_RELOCATION_RELOCATION_TIME = "relocationTime";
	public static final String LOG_RELOCATION_AFFECTED_VEHICLES = "affectedVehicles";
	public static final String LOG_RELOCATION_RELOCABLE_VEHICLES= "relocableVehicles";
	public static final String LOG_RELOCATION_SUCCES = "success";
	public static final String LOG_RELOCATION_RELOCATED_VEHICLE_ORDER = "relocatedVehicleOrder";
	public static final String LOG_RELOCATION_RELOCATED_VEHICLE_LOST_TIME = "relocatedVehicleLostTime";
	
	//Update Time 
	public static final String UPDATE_TIME = "UpdateTime";
	public static final String UPDATE_TIME_TIME_UPDATED = "timeUpdated";
	public static final String UPDATE_TIME_MINUTES_UPDATED = "minutesUpdated";


	// Ontology Instance
	private static final Ontology instance = new ParkingManagementOntology();

	// Function to access the ontology
	public static Ontology getInstance() {
		return instance;
	}

	// Constructor
	private ParkingManagementOntology() {
		// ParkingManagementOntology extends basic Ontology
		super(ONTOLOGY_NAME, BasicOntology.getInstance());

		try {

			// Add the elements
			add(new ConceptSchema(DRIVER), Driver.class);
			add(new ConceptSchema(ROAD), Road.class);
			add(new ConceptSchema(VEHICLE), Vehicle.class);
			add(new ConceptSchema(REST_AREA), Vehicle.class);
			add(new ConceptSchema(RESERVATION), Reservation.class);
			add(new AgentActionSchema(RESERVATION_REQUEST),ReservationRequest.class);
			add(new AgentActionSchema(OCCUPATION_REQUEST),OccupationRequest.class);
			add(new AgentActionSchema(CANCEL_REQUEST), CancelRequest.class);
			add(new AgentActionSchema(LOG_JOURNEY_DATA), LogJourneyData.class);
			add(new AgentActionSchema(LOG_RELOCATION), LogRelocation.class);
			add(new AgentActionSchema(UPDATE_TIME), UpdateTime.class);
			

			// Road Scheme
			ConceptSchema cs = (ConceptSchema) getSchema(ROAD);
			cs.add(ROAD_ID, (PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			cs.add(ROAD_NAME, (PrimitiveSchema) getSchema(BasicOntology.STRING));
			cs.add(ROAD_TRAFICC,(PrimitiveSchema) getSchema(BasicOntology.FLOAT));
			cs.add(ROAD_LENGHT,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));

			// Driver Scheme
			cs = (ConceptSchema) getSchema(DRIVER);
			cs.add(DRIVER_ID,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			cs.add(DRIVER_NAME,(PrimitiveSchema) getSchema(BasicOntology.STRING));
			cs.add(DRIVER_DRIVING_TIME,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));

			// Vehicle Scheme
			cs = (ConceptSchema) getSchema(VEHICLE);
			cs.add(VEHICLE_ID,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			cs.add(VEHICLE_MAXSPEED,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			cs.add(VEHICLE_ROAD, (ConceptSchema) getSchema(ROAD));
			cs.add(VEHICLE_KM,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			cs.add(VEHICLE_DRIVER, (ConceptSchema) getSchema(DRIVER));

			// Rest Area Scheme
			cs = (ConceptSchema) getSchema(REST_AREA);
			cs.add(REST_AREA_ID,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			cs.add(REST_AREA_NAME,(PrimitiveSchema) getSchema(BasicOntology.STRING));
			cs.add(REST_AREA_SPOTS,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			cs.add(REST_AREA_ROAD, (ConceptSchema) getSchema(ROAD));
			cs.add(REST_AREA_KM,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));

			// Reservation Scheme
			cs = (ConceptSchema) getSchema(RESERVATION);
			cs.add(RESERVATION_JOURNEY_MANAGER_ID,(PrimitiveSchema) getSchema(BasicOntology.STRING));
			cs.add(RESERVATION_TIME_START,(PrimitiveSchema) getSchema(BasicOntology.DATE));
			cs.add(RESERVATION_TIME_END,(PrimitiveSchema) getSchema(BasicOntology.DATE));

			// Reservation Action
			AgentActionSchema as = (AgentActionSchema) getSchema(RESERVATION_REQUEST);
			as.add(RESERVATION_REQUEST_RESERVATION,(ConceptSchema) getSchema(RESERVATION));

			// Occupation Action
			as = (AgentActionSchema) getSchema(OCCUPATION_REQUEST);
			as.add(OCCUPATION_REQUEST_RESERVATION,(ConceptSchema) getSchema(RESERVATION));

			// Cancel Action
			as = (AgentActionSchema) getSchema(CANCEL_REQUEST);
			as.add(CANCEL_REQUEST_RESERVATION,(ConceptSchema) getSchema(RESERVATION));


			//Log Journey Data
			as = (AgentActionSchema) getSchema(LOG_JOURNEY_DATA);
			as.add(LOG_JOURNEY_DATA_DRIVED_TIME,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_JOURNEY_ID_VEHICLE,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_JOURNEY_DATA_LOST_TIME,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_JOURNEY_DATA_REST_TIME,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_JOURNEY_DATA_KM,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_JOURNEY_DATA_KM_IN,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_JOURNEY_DATA_KM_OUT,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_JOURNEY_DATA_TIME_IN,(PrimitiveSchema) getSchema(BasicOntology.DATE));
			as.add(LOG_JOURNEY_DATA_TIME_OUT,(PrimitiveSchema) getSchema(BasicOntology.DATE));
			as.add(LOG_JOURNEY_INFRACTION_TYPE,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			
			
			//Log Relocation 
			as = (AgentActionSchema) getSchema(LOG_RELOCATION);
			as.add(LOG_RELOCATION_RELOCATION_TIME,(PrimitiveSchema) getSchema(BasicOntology.DATE));
			as.add(LOG_RELOCATION_AFFECTED_VEHICLES,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_RELOCATION_RELOCABLE_VEHICLES,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_RELOCATION_SUCCES,(PrimitiveSchema) getSchema(BasicOntology.BOOLEAN));
			as.add(LOG_RELOCATION_RELOCATED_VEHICLE_ORDER,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));
			as.add(LOG_RELOCATION_RELOCATED_VEHICLE_LOST_TIME,(PrimitiveSchema) getSchema(BasicOntology.FLOAT));			
				
			//Update Time 
			as = (AgentActionSchema) getSchema(UPDATE_TIME);
			as.add(UPDATE_TIME_TIME_UPDATED,(PrimitiveSchema) getSchema(BasicOntology.DATE));
			as.add(UPDATE_TIME_MINUTES_UPDATED,(PrimitiveSchema) getSchema(BasicOntology.INTEGER));

			
			
		} catch (OntologyException oe) {
			oe.printStackTrace();
		}

	}

}
